#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio
#

import sys

from typing import Optional, Type

from wagascianpy.utils.configuration import Configuration, RepositoryType

import wagascianpy.database.bsddb
import wagascianpy.database.db_record
import wagascianpy.database.wagascidb as wgdb
import wagascianpy.database.bsddb as bsddb
import wagascianpy.program.program_builder
import wagascianpy.utils.environment
import wagascianpy.utils.utils

import wagascianpy.viewer.downloader
import wagascianpy.viewer.show_run_info
import wagascianpy.viewer.topology
import wagascianpy.viewer.utils
import wagascianpy.viewer.parse_args
import wagascianpy.viewer.configuration

from wagascianpy.viewer.import_tkinter import import_messagebox, import_tkinter, import_pygubu


class Application(object):
    """Main GUI application
    """

    def __init__(self, config):
        # type: (Type[Configuration]) -> None

        # private attributes
        self._config = config
        self._batch_mode = self._config.viewer.batch_mode()
        self._records = None
        self._topology = wagascianpy.viewer.topology.Topology()

        if not self._batch_mode:
            self._build_gui()

    def _build_gui(self):

        # 0: Check that the GUI modules can be imported
        tkinter = import_tkinter()
        if not tkinter:
            raise ImportError("GUI could not be built because tkinter module was not found")
        pygubu = import_pygubu()
        if not pygubu:
            raise ImportError("pygubu module could not be imported!")

        # 1: Create a builder
        self._builder = pygubu.Builder()

        # 2: Load an gui.ui file
        self._builder.add_from_file(wagascianpy.viewer.utils.gui_file_finder())

        # 3: Create the widget using a master as parent
        self._mainwindow = self._builder.get_object('mainwindow')

        # 4: Connect method callbacks
        self._builder.connect_callbacks(self)

        # 5: override report_callback_exception method
        tkinter.Tk.report_callback_exception = self._report_callback_exception

        # 6: set default values
        self._builder.get_object('wallmrd_top_north').invoke()
        self._builder.get_object('wallmrd_bottom_north').invoke()
        self._builder.get_object('wallmrd_top_south').invoke()
        self._builder.get_object('wallmrd_bottom_south').invoke()
        self._builder.get_object('wagasci_top_upstream').invoke()
        self._builder.get_object('wagasci_side_upstream').invoke()
        self._builder.get_object('wagasci_top_downstream').invoke()
        self._builder.get_object('wagasci_side_downstream').invoke()

        # WAGASCI database configuration
        self._builder.get_object('wagasci_repository').set(self._config.wagascidb.wagasci_repository())
        self._builder.get_object('wagasci_database').set(self._config.wagascidb.wagasci_database())
        self._builder.get_object('wagasci_download_location').set(self._config.wagascidb.wagasci_download_location())
        self._builder.get_object('wagasci_decoded_location').set(self._config.wagascidb.wagasci_decoded_location())
        if self._config.wagascidb.repository_type() == RepositoryType.Simple:
            self._builder.get_object('simple_repository').invoke()
        elif self._config.wagascidb.repository_type() == RepositoryType.Borg:
            self._builder.get_object('borg_repository').invoke()
        else:
            print("Repository type not recognized : %s" % self._config.wagascidb.repository_type())

        # BSD database configuration
        self._builder.get_object('bsd_repository').set(self._config.bsddb.bsd_repository())
        self._builder.get_object('bsd_database').set(self._config.bsddb.bsd_database())
        self._builder.get_object('bsd_download_location').set(self._config.bsddb.bsd_download_location())

        # Global configuration
        self._builder.get_object('t2krun').set(self._config.global_configuration.t2krun())
        self._builder.get_object('history_location').set(self._config.global_configuration.history_location())
        self._builder.get_object('wagasci_libdir').set(self._config.global_configuration.wagasci_libdir())

        # Temperature configuration
        self._builder.get_object('temperature_sqlite_database').set(
            self._config.temperature.temperature_sqlite_database())

        # Viewer configuration
        self._builder.get_object('only_good_runs').invoke()
        self._builder.get_object('include_overlapping').invoke()

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def _report_callback_exception(self, exc, val, tb):
        """Report exceptions to the user
        """
        messagebox = import_messagebox()
        if messagebox:
            messagebox.showerror('Exception', message=str(val))
        else:
            raise val

    def _get_configuration(self):
        """Read each member
        """
        # WAGASCI database configuration
        wagasci_repository = self._builder.get_object('wagasci_repository').get()
        wagasci_database = self._builder.get_object('wagasci_database').get()
        wagasci_download_location = self._builder.get_object('wagasci_download_location').get()
        wagasci_decoded_location = self._builder.get_object('wagasci_decoded_location').get()
        if self._builder.tkvariables.__getitem__('is_borg_repo').get():
            repository_type = RepositoryType.Borg
        else:
            repository_type = RepositoryType.Simple

        self._config.wagascidb.override({
            'wagasci_repository': wagasci_repository,
            'wagasci_database': wagasci_database,
            'wagasci_download_location': wagasci_download_location,
            'wagasci_decoded_location': wagasci_decoded_location,
            'repository_type': repository_type
        })

        # BSD database configuration
        bsd_repository = self._builder.get_object('bsd_repository').get()
        bsd_database = self._builder.get_object('bsd_database').get()
        bsd_download_location = self._builder.get_object('bsd_download_location').get()

        self._config.bsddb.override({
            'bsd_repository': bsd_repository,
            'bsd_database': bsd_database,
            'bsd_download_location': bsd_download_location
        })

        # Temperature configuration
        temperature_sqlite_database = self._builder.get_object('temperature_sqlite_database').get()
        self._config.temperature.override({
            'temperature_sqlite_database': temperature_sqlite_database
        })

        # Viewer configuration
        update_wagasci_database = self._builder.tkvariables.__getitem__('update_wagasci_database').get()
        rebuild_wagasci_database = self._builder.tkvariables.__getitem__('rebuild_wagasci_database').get()
        update_bsd_database = self._builder.tkvariables.__getitem__('update_bsd_database').get()
        rebuild_bsd_database = self._builder.tkvariables.__getitem__('rebuild_bsd_database').get()
        only_good_runs = self._builder.tkvariables.__getitem__('only_good_runs').get()
        include_overlapping = self._builder.tkvariables.__getitem__('include_overlapping').get()
        self._config.viewer.override({
            'update_wagasci_database': update_wagasci_database,
            'rebuild_wagasci_database': rebuild_wagasci_database,
            'update_bsd_database': update_bsd_database,
            'rebuild_bsd_database': rebuild_bsd_database,
            'only_good_runs': only_good_runs,
            'include_overlapping': include_overlapping,
        })

        # Global configuration
        t2krun = int(self._builder.get_object('t2krun').get())
        history_location = self._builder.get_object('history_location').get()
        wagasci_libdir = self._builder.get_object('wagasci_libdir').get()
        self._config.global_configuration.override({
            't2krun': t2krun,
            'history_location': history_location,
            'wagasci_libdir': wagasci_libdir,
        })

    def _get_topology(self):
        self._topology.read_topology(self._builder)

    def _get_analyzers_list(self):
        decoder = self._builder.tkvariables.__getitem__('decoder').get()
        make_hist = self._builder.tkvariables.__getitem__('make_hist').get()
        tdc_calibration = self._builder.tkvariables.__getitem__('tdc_calibration').get()
        bcid_distribution = self._builder.tkvariables.__getitem__('bcid_distribution').get()
        adc_distribution = self._builder.tkvariables.__getitem__('adc_distribution').get()
        spill_number_fixer = self._builder.tkvariables.__getitem__('spill_number_fixer').get()
        beam_summary_data = self._builder.tkvariables.__getitem__('beam_summary_data').get()
        apply_detector_flags = self._builder.tkvariables.__getitem__('apply_detector_flags').get()
        sanity_check = self._builder.tkvariables.__getitem__('sanity_check').get()
        temperature = self._builder.tkvariables.__getitem__('temperature').get()
        adc_calibration = self._builder.tkvariables.__getitem__('adc_calibration').get()
        Configuration.analyzer_configuration.override({
            'decoder': decoder,
            'make_hist': make_hist,
            'tdc_calibration': tdc_calibration,
            'bcid_distribution': bcid_distribution,
            'adc_distribution': adc_distribution,
            'spill_number_fixer': spill_number_fixer,
            'beam_summary_data': beam_summary_data,
            'apply_detector_flags': apply_detector_flags,
            'sanity_check': sanity_check,
            'temperature': temperature,
            'adc_calibration': adc_calibration
        })

    def _database_sanity_check(self):
        wagascianpy.viewer.utils.check_repository_and_database_sanity(
            repository=self._config.wagascidb.wagasci_repository(),
            database=self._config.wagascidb.wagasci_database(),
        )
        wagascianpy.viewer.utils.check_repository_and_database_sanity(
            repository=self._config.bsddb.bsd_repository(),
            database=self._config.bsddb.bsd_database(),
        )

    def download(self, overwrite=True, builtin_reporter=True):
        # type: (bool, bool) -> None
        # Read the topology checkboxes
        if not self._batch_mode:
            self._get_topology()

        if builtin_reporter:
            reporter = self._report_callback_exception
        else:
            reporter = None

        # Download the runs
        wagascianpy.viewer.downloader.downloader(
            config=self._config,
            records=self._records,
            topology=self._topology,
            reporter=reporter,
            overwrite=overwrite,
            batch_mode=self._batch_mode)

    def update_bsd_database(self):
        bsddb.BsdDataBase(bsd_database_location=self._config.bsddb.bsd_database(),
                          bsd_repository_location=self._config.bsddb.bsd_repository(),
                          bsd_repository_download_location=self._config.bsddb.bsd_download_location(),
                          update_db=Configuration.viewer.update_bsd_database(),
                          rebuild_db=Configuration.viewer.rebuild_bsd_database(),
                          t2kruns=self._config.global_configuration.t2krun())

    def analyze(self, overwrite_flag=False):
        # type: (bool) -> None

        # 1 : update members
        if not self._batch_mode:
            self._get_configuration()
            self._get_analyzers_list()

        # 2 : Check the input folder (you need to have downloaded the raw data before decoding)
        if self._config.analyzer_configuration.decoder():
            run_dic = wagascianpy.viewer.utils.check_input_folder(
                input_folder=self._config.wagascidb.wagasci_download_location(),
                records=self._records, batch_mode=self._batch_mode)
        else:
            run_dic = wagascianpy.viewer.utils.check_input_folder(
                input_folder=self._config.wagascidb.wagasci_decoded_location(),
                records=self._records, batch_mode=self._batch_mode)

        # 3 : create builder
        builder = wagascianpy.program.program_builder.ProgramBuilder(self._config.global_configuration.wagasci_libdir())

        # 4 : set run dictionary
        builder.run_location = run_dic

        # 5 : set save location
        if self._config.analyzer_configuration.decoder() and \
                self._config.wagascidb.wagasci_decoded_location():
            builder.save_location = self._config.wagascidb.wagasci_decoded_location()
        else:
            builder.output_dir_same_as_input()

        # 6 : add decoder
        if self._config.analyzer_configuration.decoder():
            print("Selected decoder analyzer")
            builder.enforce_dependencies()
            builder.add_decoder(overwrite_flag=overwrite_flag, compatibility_mode=False)
        else:
            builder.do_not_enforce_dependencies()

        # 8 : add make hist
        if self._config.analyzer_configuration.make_hist():
            print("Selected make hist analyzer")
            builder.add_make_hist()

        # 9 : add tdc calibration
        if self._config.analyzer_configuration.tdc_calibration():
            print("Selected TDC calibration analyzer")
            builder.add_tdc_calibration()

        # 10 : add BCID distribution
        if self._config.analyzer_configuration.bcid_distribution():
            print("Selected BCID analyzer")
            builder.add_bcid_distribution(chip_by_chip=True)

        # 11 : add ADC distribution
        if self._config.analyzer_configuration.adc_distribution():
            print("Selected ADC analyzer")
            builder.add_adc_distribution(chip_by_chip=True)

        # 12 : add spill number fixer
        if self._config.analyzer_configuration.spill_number_fixer():
            print("Selected spill number fixer analyzer")
            builder.add_spill_number_fixer(enable_graphics=False)

        # 13 : add beam summary data
        if self._config.analyzer_configuration.beam_summary_data():
            print("Selected BSD analyzer")
            builder.add_beam_summary_data(
                bsd_database_location=self._config.bsddb.bsd_database(),
                bsd_repository_location=self._config.bsddb.bsd_repository(),
                download_bsd_database_location='/tmp/bsddb.db',
                download_bsd_repository_location=self._config.bsddb.bsd_download_location(),
                t2krun=self._config.global_configuration.t2krun(),
                recursive=True)

        # 14 : add detector flags
        if self._config.analyzer_configuration.apply_detector_flags():
            print("Selected apply detector flags analyzer")
            builder.add_apply_detector_flags(wagasci_database=self._config.wagascidb.wagasci_database())

        # 14 : add temperature
        if self._config.analyzer_configuration.temperature():
            print("Selected temperature analyzer")
            builder.add_temperature(sqlite_database=self._config.temperature.temperature_sqlite_database())

        # 15 : add ADC calibration
        if self._config.analyzer_configuration.adc_calibration():
            print("Selected ADC calibration analyzer")
            builder.multiple_runs_analyzer_save_location = self._config.global_configuration.history_location()
            builder.add_adc_calibration()

        # 16 : add sanity check
        if self._config.analyzer_configuration.sanity_check():
            print("Selected sanity check analyzer")
            builder.add_sanity_check(wagasci_database=self._config.wagascidb.wagasci_database())

        # 16 : start program
        builder.program.start()

    def get_time_interval(self, start_time=None, stop_time=None, only_good_runs=True, include_overlapping=True):
        # type: (Optional[str], Optional[str], bool, bool) -> None
        """Get runs in an interval of time
        """

        if not self._batch_mode:
            self._get_configuration()
            only_good_runs = self._builder.tkvariables.__getitem__('only_good_runs').get()
            include_overlapping = self._builder.tkvariables.__getitem__('include_overlapping').get()
            start_time, stop_time = wagascianpy.viewer.utils.get_time_interval_members(self._builder)
        self._database_sanity_check()

        with wgdb.WagasciDataBase(db_location=self._config.wagascidb.wagasci_database(),
                                  repo_location=self._config.wagascidb.wagasci_repository(),
                                  is_borg_repo=self._config.wagascidb.repository_type() == RepositoryType.Borg,
                                  update_db=Configuration.viewer.update_wagasci_database(),
                                  rebuild_db=Configuration.viewer.rebuild_wagasci_database(),
                                  wagasci_libdir=self._config.global_configuration.wagasci_libdir()) as database:
            self._records = database.get_time_interval(datetime_start=start_time,
                                                       datetime_stop=stop_time,
                                                       only_good=only_good_runs,
                                                       include_overlapping=include_overlapping)

            if not self._batch_mode:
                wagascianpy.viewer.show_run_info.make_records_table(config=self._config,
                                                                    builder=self._builder,
                                                                    records=self._records)
            else:
                for record in self._records:
                    record.pretty_print()

    def get_run_interval(self, start_run=None, stop_run=None, only_good_runs=True):
        # type: (Optional[int], Optional[int], bool) -> None
        """Get runs in an between two runs
        """
        if not self._batch_mode:
            self._get_configuration()
            only_good_runs = self._builder.tkvariables.__getitem__('only_good_runs').get()
            start_run = self._builder.tkvariables.__getitem__('start_run').get()
            stop_run = self._builder.tkvariables.__getitem__('stop_run').get()
        assert start_run is not None and stop_run is not None, "Please specify start run and stop run"
        self._database_sanity_check()

        with wgdb.WagasciDataBase(db_location=self._config.wagascidb.wagasci_database(),
                                  repo_location=self._config.wagascidb.wagasci_repository(),
                                  is_borg_repo=self._config.wagascidb.repository_type() == RepositoryType.Borg,
                                  update_db=Configuration.viewer.update_wagasci_database(),
                                  rebuild_db=Configuration.viewer.rebuild_wagasci_database(),
                                  wagasci_libdir=self._config.global_configuration.wagasci_libdir()) as database:
            self._records = database.get_run_interval(run_number_start=start_run,
                                                      run_number_stop=stop_run,
                                                      only_good=only_good_runs)

            if not self._batch_mode:
                wagascianpy.viewer.show_run_info.make_records_table(config=self._config,
                                                                    builder=self._builder,
                                                                    records=self._records)
            else:
                for record in self._records:
                    record.pretty_print()

    def get_all(self, only_good_runs=True):
        # type: (bool) -> None
        """Get all WAGASCI runs
        """
        if not self._batch_mode:
            self._get_configuration()
            only_good_runs = self._builder.tkvariables.__getitem__('only_good_runs').get()
        self._database_sanity_check()

        with wgdb.WagasciDataBase(db_location=self._config.wagascidb.wagasci_database(),
                                  repo_location=self._config.wagascidb.wagasci_repository(),
                                  is_borg_repo=self._config.wagascidb.repository_type() == RepositoryType.Borg,
                                  update_db=self._config.viewer.update_wagasci_database(),
                                  rebuild_db=self._config.viewer.rebuild_wagasci_database(),
                                  wagasci_libdir=self._config.global_configuration.wagasci_libdir()) as database:
            self._records = database.get_all(only_good=only_good_runs)

            if not self._batch_mode:
                wagascianpy.viewer.show_run_info.make_records_table(config=self._config,
                                                                    builder=self._builder,
                                                                    records=self._records)
            else:
                for record in self._records:
                    record.pretty_print()

    def run(self):
        """Run the application
        """
        if not self._batch_mode:
            self._mainwindow.mainloop()


if __name__ == '__main__':

    # Parse shell arguments
    ARGS = wagascianpy.viewer.parse_args.parse_args(sys.argv[1:])

    # Edit the initial configuration
    wagascianpy.viewer.configuration.fill_configuration(ARGS)

    # Create application
    APP = Application(config=Configuration)

    if Configuration.viewer.batch_mode():
        # select runs
        if Configuration.run_select.get_time_interval():
            APP.get_time_interval(start_time=Configuration.run_select.start_time(),
                                  stop_time=Configuration.run_select.stop_time(),
                                  only_good_runs=Configuration.viewer.only_good_runs(),
                                  include_overlapping=Configuration.viewer.include_overlapping())
        elif Configuration.run_select.get_run_interval():
            APP.get_run_interval(start_run=Configuration.run_select.start_run(),
                                 stop_run=Configuration.run_select.stop_run(),
                                 only_good_runs=Configuration.viewer.only_good_runs())
        elif Configuration.run_select.get_all():
            APP.get_all(only_good_runs=Configuration.viewer.only_good_runs())

        # download
        if Configuration.analyzer_configuration.download():
            APP.download(builtin_reporter=False)

        # Update or rebuild BSD database
        if Configuration.viewer.rebuild_bsd_database() or Configuration.viewer.update_bsd_database():
            APP.update_bsd_database()

        # analysis
        APP.analyze(overwrite_flag=Configuration.analyzer_configuration.overwrite_flag())

    else:
        APP.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio
#
import argparse
import os
import textwrap
from datetime import datetime, timedelta, date, time
from typing import Generator, List

import ROOT

import wagascianpy.database.wagascidb as db
import wagascianpy.utils.utils
import wagascianpy.utils.treenames as treenames
from wagascianpy.analysis.spill import WAGASCI_SPILL_BEAM_MODE
from wagascianpy.database.db_record import DBRecord as DBr
from wagascianpy.plotting.detector import Detectors

ROOT.PyConfig.IgnoreCommandLineOptions = True


def _daterange(start_date, end_date):
    # type: (date, date) -> Generator[date]
    for n in range(int((end_date - start_date).days) + 1):
        yield start_date + timedelta(n)


def _decoded_path(root_dir, file, runname, dif_id):
    # type: (str, str, str, int) -> str
    decoded_filename = None
    filename, file_extension = os.path.splitext(file)
    if file_extension == '.raw':
        decoded_filename = "{}_ecal_dif_{}_tree.root".format(runname, dif_id)
    elif file_extension == '.root':
        decoded_filename = file
    return os.path.join(root_dir, runname, decoded_filename)


def _get_active_detectors(runs, root_dir):
    # type: (List[db.WagasciRunRecord], str) -> Detectors
    detectors = Detectors()
    detectors.disable_all_difs()
    for one_run in runs:
        for dif_id, file in one_run.raw_files.items():
            decoded_fpath = _decoded_path(root_dir=root_dir, file=file, runname=one_run.name, dif_id=dif_id)
            print("Day {}: DIF {} is enabled".format(day.strftime("%Y-%m-%d"), dif_id))
            print("Adding DIF {} decoded file : {}".format(dif_id, decoded_fpath))
            detectors.enable_dif(dif_id)
            detectors.get_dif(dif_id).add_tree(root_file=decoded_fpath,
                                               tree_name=wagascianpy.utils.utils.extract_raw_tree_name(decoded_fpath),
                                               tree_friends=[treenames.BSD_TREE_NAME, treenames.FSN_TREE_NAME,
                                                             treenames.TEMPHUM_TREE_NAME, treenames.PEU_TREE_NAME,
                                                             treenames.TIME_NS_TREE_NAME, treenames.FLAGS_TREE_NAME])
    return detectors


def _clone_tree(input_tree, output_tfile):
    # type: (str, ROOT.TFile) -> ROOT.TChain
    output_tree = input_tree.CloneTree(0)
    output_tree.SetDirectory(output_tfile)
    output_tree.SetAutoFlush(100)
    output_tree.SetAutoSave(100)
    return output_tree


if __name__ == "__main__":
    parser = argparse.ArgumentParser(usage='use "python %(prog)s --help" for more information',
                                     argument_default=None, description=textwrap.dedent('''\
                                     Split the decoded runs into one folder per each day.'''))

    parser.add_argument('-d', '--wagasci-database', metavar='<WAGASCI run database>', dest='wagasci_database',
                        type=str, required=True, help=textwrap.dedent('''\
                        Path to the WAGASCI run database file. This file is created by the 
                        Wagasci Database Viewer (wagasciviewer.py) program and is usually called wagascidb.db.
                        '''))

    parser.add_argument('-f', '--wagasci-decoded', metavar='<WAGASCI runs directory>',
                        dest='wagasci_decoded', type=str, required=True, help=textwrap.dedent('''\
                        Path to the directory containing all the WAGASCI decoded runs.'''))

    parser.add_argument('-o', '--output-directory', metavar='<output directory>', dest='output_directory',
                        type=str, required=True, help=textwrap.dedent('''\
                        Directory where to store all the output files
                        '''))

    parser.add_argument('-a', '--start-time', metavar='<start time>', dest='start_time', type=str,
                        help='Start date and time in the format "%%Y/%%m/%%d %%H:%%M:%%S"', default=None,
                        required=False)

    parser.add_argument('-b', '--stop-time', metavar='<stop time>', dest='stop_time', type=str,
                        help='Stop date and time in the format "%%Y/%%m/%%d %%H:%%M:%%S"', default=None,
                        required=False)

    parser.add_argument('-m', '--start-run', metavar='<start run>', dest='start_run', type=int,
                        help='start run number', default=None, required=False)

    parser.add_argument('-n', '--stop-run', metavar='<stop run>', dest='stop_run', type=int,
                        help='Stop run number', default=None, required=False)

    parser.add_argument('-t', '--t2krun', metavar='<T2K run>', dest='t2krun', type=int,
                        help='T2K run number (default is 10)', default=10, required=False)

    parser.add_argument('-g', '--only-good', dest='only_good', required=False, default=True,
                        action='store_true', help=textwrap.dedent('''\
                        Only select runs and detectors whose flag is set to good (good run or good data)'''))

    parser.add_argument('-c', '--day-counter-offset', metavar='<offset>', type=int, dest='day_counter_offset',
                        required=False, default=0, help=textwrap.dedent('''\
                        Offset to be added to the counter that counts the number of days. The day counter is used 
                        as a placeholder for the run number. It is useful to give an offset if you are running this 
                        script multiple times and want to keep the run number consecutive.'''))

    arguments = parser.parse_args()

    time_interval = bool(arguments.start_time) and bool(arguments.stop_time)
    run_interval = bool(arguments.start_run) and bool(arguments.stop_run)

    if time_interval and not run_interval:
        start = DBr.str2datetime(arguments.start_time)
        stop = DBr.str2datetime(arguments.stop_time)
    elif run_interval and not time_interval:
        start = arguments.start_run
        stop = arguments.stop_run
    else:
        raise ValueError("Please specify (--start-time AND --stop-time) OR (--start-run AND --stop-run)")

    # If the run interval is given extract the start and stop time (if the start and stop time is already set this
    # function does nothing)
    start, stop = db.run_to_interval(start=start, stop=stop, database=arguments.wagasci_database)
    print("Start time : {}".format(start.strftime("%Y-%m-%d %H:%M:%S")))
    print("Stop time : {}".format(stop.strftime("%Y-%m-%d %H:%M:%S")))

    # for each day in the range
    for day_counter, day in enumerate(_daterange(start_date=start.date(), end_date=stop.date())):
        print("Splitting day {}".format(day.strftime("%Y-%m-%d")))

        day_begin = DBr.datetime2timestamp(datetime.combine(
            date=day, time=time(hour=0, minute=0, second=0, microsecond=0)))
        day_end = DBr.datetime2timestamp(datetime.combine(
            date=day, time=time(hour=23, minute=59, second=59, microsecond=999999)))

        with db.WagasciDataBase(db_location=arguments.wagasci_database, repo_location=arguments.wagasci_decoded,
                                is_borg_repo=False) as database:
            daily_runs = database.get_time_interval(datetime_start=day_begin,
                                                    datetime_stop=day_end,
                                                    only_good=arguments.only_good,
                                                    include_overlapping=True)
            if not daily_runs:
                print("No run found in day {}".format(day.strftime("%Y-%m-%d")))
                continue

            print("Found {} runs in day {}".format(len(daily_runs), day.strftime("%Y-%m-%d")))
            for i, run in enumerate(daily_runs):
                print("  Run {}: {}".format(i, run.name))

            active_detectors = _get_active_detectors(runs=daily_runs, root_dir=arguments.wagasci_decoded)

            if not active_detectors:
                print("No active detector found in day {}".format(day.strftime("%Y-%m-%d")))
                continue

            output_directory = "physics_run_{}".format(day.strftime("%Y-%m-%d"))

            for detector in active_detectors:
                for dif in [dif for dif in detector if dif.is_enabled()]:
                    print("Day {}: Saving DIF {} data".format(day.strftime("%Y-%m-%d"), dif.dif_id))
                    output_runname = "physics_run_{}_00-00-00_{}".format(day.strftime("%Y-%m-%d"),
                                                                         arguments.day_counter_offset + day_counter)
                    output_filename = "{}_ecal_dif_{}_tree.root".format(output_runname, dif.dif_id)
                    output_directory = os.path.join(arguments.output_directory, output_runname)
                    if not os.path.exists(output_directory):
                        os.makedirs(output_directory, mode=0o775)
                    output_path = os.path.join(output_directory, output_filename)

                    tfile = ROOT.TFile(output_path, "RECREATE")
                    raw = _clone_tree(input_tree=dif.get_tree(), output_tfile=tfile)

                    tree_friends = {treenames.BSD_TREE_NAME: None,
                                    treenames.FSN_TREE_NAME: None,
                                    treenames.TEMPHUM_TREE_NAME: None,
                                    treenames.PEU_TREE_NAME: None,
                                    treenames.TIME_NS_TREE_NAME: None,
                                    treenames.FLAGS_TREE_NAME: None}
                    tree_clones = tree_friends.copy()
                    for friend in tree_friends.keys():
                        if dif.has_friend(friend):
                            tree_friends[friend] = dif.get_friend(friend)
                    for friend, tree in tree_friends.items():
                        if tree is not None:
                            tree_clones[friend] = _clone_tree(input_tree=dif.get_friend(friend), output_tfile=tfile)
                    for j, event in enumerate(dif.get_tree()):
                        if j % 1000 == 0:
                            print("Day {}: DIF {}: processed {} events".format(day.strftime("%Y-%m-%d"), dif.dif_id, j))
                        if event.spill_mode == WAGASCI_SPILL_BEAM_MODE and day_begin < event.timestamp < day_end:
                            raw.Fill()
                            for tree in tree_clones.values():
                                if tree is not None:
                                    tree.Fill()



import unittest

import wagascianpy.plotting.topology


class TestTopologyIterateByDetector(unittest.TestCase):

    def setUp(self):
        self.topology = wagascianpy.plotting.topology.Topology(iterate_by_dif=False)

    def test_are_all_enabled(self):
        self.assertTrue(self.topology.are_all_enabled())
        self.topology.wagasci_downstream.disable()
        self.assertFalse(self.topology.are_all_enabled())

    def test_disable_all(self):
        self.assertTrue(self.topology.are_all_enabled())
        self.topology.disable_all()
        self.assertFalse(self.topology.are_all_enabled())
        self.topology.wallmrd_north.enable()
        self.assertTrue(self.topology.wallmrd_north.is_enabled())
        self.assertFalse(self.topology.wallmrd_south.is_enabled())
        self.assertFalse(self.topology.wagasci_upstream.is_enabled())
        self.assertFalse(self.topology.wagasci_downstream.is_enabled())

    def test_disable_all_but_one(self):
        self.assertTrue(self.topology.are_all_enabled())
        self.topology.disable_all_but("WAGASCI downstream")
        self.assertFalse(self.topology.are_all_enabled())
        self.assertTrue(self.topology.wagasci_downstream.is_enabled())
        self.assertFalse(self.topology.wagasci_upstream.is_enabled())
        self.assertFalse(self.topology.wallmrd_south.is_enabled())
        self.assertFalse(self.topology.wallmrd_north.is_enabled())

    def test_get_all(self):
        self.assertEqual(4, len(self.topology.get_all()))
        self.assertIn(self.topology.wagasci_upstream, self.topology.get_all())
        self.assertIn(self.topology.wagasci_downstream, self.topology.get_all())
        self.assertIn(self.topology.wallmrd_north, self.topology.get_all())
        self.assertIn(self.topology.wallmrd_south, self.topology.get_all())

    def test_get_disabled(self):
        self.topology.wagasci_downstream.disable()
        self.assertEqual(1, len(self.topology.get_disabled()))
        self.assertIn(self.topology.wagasci_downstream, self.topology.get_disabled())

    def test_get_enabled(self):
        self.topology.wagasci_downstream.disable()
        self.assertEqual(3, len(self.topology.get_enabled()))
        self.assertIn(self.topology.wagasci_upstream, self.topology.get_enabled())
        self.assertIn(self.topology.wallmrd_north, self.topology.get_enabled())
        self.assertIn(self.topology.wallmrd_south, self.topology.get_enabled())

    def test_import_topology(self):
        topology = wagascianpy.plotting.topology.Topology(iterate_by_dif=True)
        topology.wagasci_downstream_top.disable()
        topology.wagasci_downstream_side.disable()
        self.topology.import_topology(topology)
        self.assertFalse(self.topology.wagasci_downstream.is_enabled())


class TestTopologyIterateByDif(unittest.TestCase):

    def setUp(self):
        self.topology = wagascianpy.plotting.topology.Topology(iterate_by_dif=True)

    def test_are_all_enabled(self):
        self.assertTrue(self.topology.are_all_enabled())
        self.topology.wagasci_downstream_side.disable()
        self.assertFalse(self.topology.are_all_enabled())

    def test_disable_all(self):
        self.assertTrue(self.topology.are_all_enabled())
        self.topology.disable_all()
        self.assertFalse(self.topology.are_all_enabled())

    def test_disable_all_but_one(self):
        self.assertTrue(self.topology.are_all_enabled())
        self.topology.disable_all_but("WAGASCI downstream side")
        self.assertFalse(self.topology.are_all_enabled())
        self.assertFalse(self.topology.wagasci_downstream_top.is_enabled())
        self.assertTrue(self.topology.wagasci_downstream_side.is_enabled())

    def test_get_all(self):
        self.assertEqual(8, len(self.topology.get_all()))
        self.assertIn(self.topology.wagasci_upstream_top, self.topology.get_all())
        self.assertIn(self.topology.wagasci_downstream_top, self.topology.get_all())
        self.assertIn(self.topology.wallmrd_north_top, self.topology.get_all())
        self.assertIn(self.topology.wallmrd_south_top, self.topology.get_all())
        self.assertIn(self.topology.wagasci_upstream_side, self.topology.get_all())
        self.assertIn(self.topology.wagasci_downstream_side, self.topology.get_all())
        self.assertIn(self.topology.wallmrd_north_bottom, self.topology.get_all())
        self.assertIn(self.topology.wallmrd_south_bottom, self.topology.get_all())

    def test_get_disabled(self):
        self.topology.wagasci_downstream_side.disable()
        self.assertEqual(1, len(self.topology.get_disabled()))
        self.assertFalse(self.topology.wagasci_downstream_side.is_enabled())
        self.assertTrue(self.topology.wagasci_downstream_top.is_enabled())
        self.assertTrue(self.topology.wagasci_upstream_top.is_enabled())
        self.assertTrue(self.topology.wagasci_upstream_side.is_enabled())
        self.assertTrue(self.topology.wallmrd_north_top.is_enabled())
        self.assertTrue(self.topology.wallmrd_north_bottom.is_enabled())
        self.assertTrue(self.topology.wallmrd_south_top.is_enabled())
        self.assertTrue(self.topology.wallmrd_south_bottom.is_enabled())
        self.assertIn(self.topology.wagasci_downstream_side, self.topology.get_disabled())

    def test_get_enabled(self):
        self.topology.wagasci_downstream_side.disable()
        self.assertEqual(7, len(self.topology.get_enabled()))
        self.assertIn(self.topology.wagasci_upstream_top, self.topology.get_enabled())
        self.assertIn(self.topology.wagasci_downstream_top, self.topology.get_enabled())
        self.assertIn(self.topology.wallmrd_north_top, self.topology.get_enabled())
        self.assertIn(self.topology.wallmrd_south_top, self.topology.get_enabled())
        self.assertIn(self.topology.wagasci_upstream_side, self.topology.get_enabled())
        self.assertNotIn(self.topology.wagasci_downstream_side, self.topology.get_enabled())
        self.assertIn(self.topology.wallmrd_north_bottom, self.topology.get_enabled())
        self.assertIn(self.topology.wallmrd_south_bottom, self.topology.get_enabled())

    def test_import_topology(self):
        topology = wagascianpy.plotting.topology.Topology(iterate_by_dif=False)
        topology.wagasci_downstream.disable()
        self.topology.import_topology(topology)
        self.assertFalse(self.topology.wagasci_downstream_top.is_enabled())
        self.assertFalse(self.topology.wagasci_downstream_side.is_enabled())


class TestDifIndex(unittest.TestCase):

    def test_dif_index_true(self):
        self.assertTrue(wagascianpy.plotting.topology.DifIndex.is_wallmrd_north(0))
        self.assertTrue(wagascianpy.plotting.topology.DifIndex.is_wallmrd_south(3))
        self.assertTrue(wagascianpy.plotting.topology.DifIndex.is_wallmrd(2))
        self.assertTrue(wagascianpy.plotting.topology.DifIndex.is_wagasci_upstream(5))
        self.assertTrue(wagascianpy.plotting.topology.DifIndex.is_wagasci_downstream(6))
        self.assertTrue(wagascianpy.plotting.topology.DifIndex.is_wagasci(4))

    def test_dif_index_false(self):
        self.assertFalse(wagascianpy.plotting.topology.DifIndex.is_wallmrd_north(6))
        self.assertFalse(wagascianpy.plotting.topology.DifIndex.is_wallmrd_south(4))
        self.assertFalse(wagascianpy.plotting.topology.DifIndex.is_wallmrd(10))
        self.assertFalse(wagascianpy.plotting.topology.DifIndex.is_wagasci_upstream(2))
        self.assertFalse(wagascianpy.plotting.topology.DifIndex.is_wagasci_downstream(3))
        self.assertFalse(wagascianpy.plotting.topology.DifIndex.is_wagasci(1))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2019 Pintaudi Giorgio

import collections
import errno
import hashlib
import inspect
import json
import os
import random
import re
import shutil
import subprocess
import sys
from collections import OrderedDict
from typing import Optional, Dict, Any, Union, List, Tuple

import paramiko
from scp import SCPClient
from six import string_types
from undecorated import undecorated

import wagascianpy.plotting
import wagascianpy.utils.environment
from wagascianpy.plotting.detector import Detectors
from wagascianpy.utils.treenames import RAW_TREE_NAME

try:
    import ROOT

    ROOT.PyConfig.IgnoreCommandLineOptions = True
except ImportError:
    ROOT = False

try:
    # noinspection PyCompatibility
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

_DEFAULT_TIMEZONE = "Asia/Tokyo"


###############################################################################
#                                 copytree                                    #
###############################################################################

def copytree(src, dst, dif_topology=None, symlinks=False, ignore=None):
    """Copy run_root_dir folder into dst folder (it does not create the dst folder
    if it does not exist)

    """
    if os.path.exists(src) and os.path.isdir(src):
        if not os.listdir(src):
            raise ValueError('Directory "%s" is empty' % src)
    else:
        raise ValueError('Directory "%s" doesn\'t exist' % src)

    dif_list = []
    dif_counter = 0
    if dif_topology is not None:
        dif_list = sorted(list(dif_topology.keys()))
    for item in sorted(os.listdir(src)):
        source_file = os.path.join(src, item)
        destination_file = os.path.join(dst, item)
        if os.path.isdir(source_file):
            shutil.copytree(source_file, destination_file, symlinks, ignore)
        else:
            if dif_topology is not None and "ecal_dif" in source_file:
                destination_file = re.sub(r"_ecal_dif_[0-9]+.raw",
                                          "_ecal_dif_%s.raw" % (str(dif_list[dif_counter])),
                                          destination_file)
                dif_counter += 1
            shutil.copy2(source_file, destination_file)


###############################################################################
#                               renametree                                    #
###############################################################################

def renametree(run_root_dir, run_name, dif_topology):
    # type:(str, str, Dict) -> None
    """
    Rename runs in folder according to topology
    :param run_root_dir: Folder where the raw data files are stored
    :param run_name: name of the run
    :param dif_topology: topology dictionary
    :return: None
    """
    if not isinstance(dif_topology, dict):
        raise ValueError("DIF topology must be a dictionary!")

    if os.path.exists(run_root_dir) and os.path.isdir(run_root_dir):
        if not os.listdir(run_root_dir):
            raise ValueError('Directory "%s" is empty' % run_root_dir)
    else:
        raise ValueError('Directory "%s" doesn\'t exist' % run_root_dir)
    topology_dif_list = sorted(list(dif_topology.keys()))
    file_dif_list = []
    for file_name in sorted(wagascianpy.utils.utils.find_files(path=run_root_dir, extension='raw')):
        if "ecal_dif" in file_name:
            match = re.search(r"_ecal_dif_([0-9]+).raw", file_name)
            if match is not None:
                file_dif_list.append(match.group(1))
    if collections.Counter(topology_dif_list) == collections.Counter(file_dif_list):
        return
    if len(file_dif_list) != len(topology_dif_list):
        raise ValueError("Number of raw files ({}) is different from topology ({})".format(len(file_dif_list),
                                                                                           len(topology_dif_list)))
    dictionary = dict(zip(sorted(file_dif_list), sorted(topology_dif_list)))
    for file_dif_id, topology_dif_id in dictionary.items():
        old_file_name = "{}_ecal_dif_{}.raw".format(run_name, file_dif_id)
        tmp_file_name = "{}_ecal_dif_{}.raw.tmp".format(run_name, topology_dif_id)
        os.rename(os.path.join(run_root_dir, old_file_name), os.path.join(run_root_dir, tmp_file_name))
    for tmp_file_name in sorted(wagascianpy.utils.utils.find_files(path=run_root_dir, extension='tmp')):
        new_file_name = tmp_file_name.replace('.tmp', '')
        os.rename(os.path.join(run_root_dir, tmp_file_name), os.path.join(run_root_dir, new_file_name))


###############################################################################
#                              static_vars                                    #
###############################################################################

def static_vars(**kwargs):
    """Static variables inside function for Python"""

    def decorate(func):
        """Decorator for static_vars"""
        for k in kwargs:
            setattr(func, k, kwargs[k])
        return func

    return decorate


###############################################################################
#                                  mkdir_p                                    #
###############################################################################

def mkdir_p(path):
    """Create a directory recursively. If the directory already exists, do
    nothing

    """
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


###############################################################################
#                                 join_threads                                #
###############################################################################

def join_threads(threads):
    """ Join all the threads in the list *threads*"""
    for thread in threads:
        result = thread.join()
        if result not in [0, -1]:
            print("Error number : %s" % result)
            exit(result)


###############################################################################
#                                limit_threads                                #
###############################################################################

def limit_threads(threads, max_threads):
    """Limit the number of simultaneuously running threads"""
    while len(threads) > max_threads:
        result = threads[0].join()
        if result not in [0, -1]:
            print("Error number : %s" % result)
            exit(result)
        threads.pop(0)


###############################################################################
#                              join_single_chain                              #
###############################################################################

def join_single_chain(chain_for_each_dif):
    """ Join all chains of threads for a single run """
    for dif_id, chain in chain_for_each_dif.items():
        print("JOIN DIF {} : LINK ID {} : THREAD ID {}".format(dif_id, id(chain.link), id(chain.thread)))
        result = chain.thread.join()
        if result not in [0, -1]:
            raise RuntimeError("DIF {} : Error code {}".format(dif_id, result))


###############################################################################
#                                 join_chains                                 #
###############################################################################

def join_chains(chains_for_each_run):
    """ Join all chains of threads in the list of chains """
    for run_name, chain_for_each_dif in chains_for_each_run.items():
        for dif_id, chain in chain_for_each_dif.items():
            print("JOIN DIF {} : LINK ID {} : THREAD ID {}".format(dif_id, id(chain.link), id(chain.thread)))
            result = chain.thread.join()
            if result not in [0, -1]:
                raise RuntimeError("RUN {} DIF {} : Error code {}".format(run_name, dif_id, result))


###############################################################################
#                                count_threads                                #
###############################################################################

def count_threads(run_chains):
    """Count total number of active threads in a chain"""
    num_threads = 0
    for dif_chains in run_chains.values():
        for chain in dif_chains.values():
            if chain.thread.is_alive():
                num_threads += 1
    return num_threads


###############################################################################
#                                limit_chains                                 #
###############################################################################

def limit_chains(run_chains, max_threads):
    """Limit the number of simultaneuously running threads"""
    num_threads = count_threads(run_chains)
    for run_name, dif_chains in sorted(run_chains.items()):
        for dif_id, chain in sorted(dif_chains.items()):
            if num_threads <= max_threads:
                return
            if chain.thread.is_alive():
                result = chain.thread.join()
                num_threads -= 1
                if result not in [0, -1]:
                    raise RuntimeError("RUN {} DIF {} : Error code {}".format(run_name, dif_id, result))


###############################################################################
#                            list_dir_with_integer                            #
###############################################################################

def list_dir_with_integer(path):
    # type: (str) -> List
    """List all directories containing an integer.

    Args:
        path: path to a directories containing other directories

    Return:
        list of tuples named 'dir_with_integer'. Each tuple contains
        two elements:
            - num eger
            - num_dir : directory name

    """
    if not os.path.exists(path) or not os.path.isdir(path):
        raise ValueError("Directory does not exists : %s" % path)
    dir_list = os.listdir(path)
    if not dir_list:
        raise ValueError("Empty directory : %s" % path)
    nums_re = re.compile(r"\d+")
    dir_with_integer = collections.namedtuple('dir_with_integer', 'num num_dir')
    result = []
    for directory in dir_list:
        if bool(nums_re.search(directory)):
            result.append(dir_with_integer(nums_re.search(directory).group(0), directory))
    return sorted(result)


###############################################################################
#                         get_immediate_subdirectories                        #
###############################################################################

def get_immediate_subdirectories(a_dir):
    return [name for name in os.listdir(a_dir)
            if os.path.isdir(os.path.join(a_dir, name))]


###############################################################################
#                              find_first_match                               #
###############################################################################

def find_first_match(name, path):
    for root, dirs, files in os.walk(path):
        if name in files:
            return os.path.join(root, name)


###############################################################################
#                              find_files_with_ext                            #
###############################################################################

def find_files(path, extension, recursive=True):
    file_list = []
    if recursive:
        for root, _, files in os.walk(path):
            for f in files:
                if f.lower().endswith('.%s' % extension.strip('.')):
                    file_list.append(f if os.path.isabs(f) else os.path.join(root, f))
    else:
        for f in [f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]:
            if f.lower().endswith('.%s' % extension.strip('.')):
                file_list.append(f if os.path.isabs(f) else os.path.join(path, f))
    return sorted(file_list)


###############################################################################
#                                extract_dif_id                               #
###############################################################################


def extract_dif_id(path):
    # type: (str) -> Optional[int]
    if not path:
        return None
    if isinstance(path, list):
        path = path[0]
    if isinstance(path, string_types) and ',' in path:
        path = path.split(',')[0]
    file_name = os.path.basename(path)
    try:
        dif_id = int(re.search(r'_dif_([\d]+)', file_name).group(1))
    except (IndexError, AttributeError):
        dif_id = None
    return dif_id


###############################################################################
#                            extract_raw_tree_name                            #
###############################################################################

def extract_raw_tree_name(tfile):
    # type: (Union[str, ROOT.TFile]) -> str
    """
    Find the name of the raw data TTree. The raw data TTRee name changed in the past
    so this function is here just for compatibility reasons. The new name is "raw".
    :param tfile: TFile where the TTree resides. Can be a string or a TFile object
    :return: name of the raw data TTree
    """
    if ROOT:
        if isinstance(tfile, string_types):
            tfile = ROOT.TFile(tfile)
        if tfile.GetListOfKeys().Contains(RAW_TREE_NAME):
            return RAW_TREE_NAME
        else:
            tree_name = "tree_dif_%s" % extract_dif_id(tfile.GetName())
            if tfile.GetListOfKeys().Contains(tree_name):
                return tree_name
            else:
                raise AttributeError("TTree not found in TFile %s" % tfile.GetName())
    else:
        raise ImportError("ROOT module not found")


###############################################################################
#                              extract_user_info                              #
###############################################################################

def extract_user_info(tfile, info):
    # type: (Union[str, ROOT.TFile], str) -> Any
    """
    Extract the user info from a raw data TTree inside a TFile
    :param tfile: TFile where the TTree resides. Can be a string or a TFile object
    :param info: name of the parameter to extract
    :return: parameter value
    """
    if ROOT:
        if isinstance(tfile, string_types):
            tfile = ROOT.TFile(tfile)
        treename = extract_raw_tree_name(tfile)
        tree = getattr(tfile, treename)
        if tree.GetUserInfo().FindObject(info):
            return tree.GetUserInfo().FindObject(info).GetVal()
        else:
            raise KeyError("info {} not found in TFile".format(info))
    else:
        raise ImportError("ROOT module not found")


###############################################################################
#                            change_directory (Cd)                            #
###############################################################################

class Cd(object):
    """Context manager for changing the current working directory"""

    def __init__(self, new_path):
        self.saved_path = None
        self.new_path = os.path.expanduser(new_path)

    def __enter__(self):
        self.saved_path = os.getcwd()
        os.chdir(self.new_path)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.saved_path)


###############################################################################
#                                    which                                    #
###############################################################################

def which(program):
    """ python implementation of the which bash command"""

    def _is_exe(file_path):
        """Check if file is executable or not"""
        return os.path.isfile(file_path) and os.access(file_path, os.X_OK)

    fpath, _ = os.path.split(program)
    if fpath:
        if _is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if _is_exe(exe_file):
                return exe_file
    return None


###############################################################################
#                                 run_cmd                                     #
###############################################################################

def run_cmd(cmd):
    """ Run a shell command"""

    try:
        stdout = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as exception:
        if exception.output:
            raise RuntimeError("Command cmd failed with outout : %s" % str(exception.output))
        raise RuntimeError("failed to execute command %s" % (str(cmd)))

    return str(stdout.decode())


###############################################################################
#                              run_borg_cmd                                   #
###############################################################################

def run_borg_cmd(cmdlist, check_progress=False):
    """ Run a borg command"""

    # Uncomment to debug borg
    # print(cmdlist)

    try:
        environment = dict(os.environ)
        environment["BORG_UNKNOWN_UNENCRYPTED_REPO_ACCESS_IS_OK"] = "yes"
        if check_progress:
            return subprocess.Popen(cmdlist,
                                    shell=True,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    env=environment)
        else:
            stdout = subprocess.check_output(cmdlist,
                                             shell=True,
                                             stderr=subprocess.STDOUT,
                                             env=environment)

    except subprocess.CalledProcessError as exception:
        if exception.output:
            print(exception.output)
            for line in exception.output.splitlines():
                json_out = json.loads(line)
                if json_out['levelname'] in [u'WARNING', u'ERROR', u'CRITICAL']:
                    raise RuntimeError(json_out['message'])
        raise RuntimeError("failed to execute command %s" % (str(cmdlist)))

    return str(stdout.decode())


###############################################################################
#                             create_ssh_client                               #
###############################################################################

def _create_ssh_client(hostname):
    client = paramiko.SSHClient()
    paramiko.transport.banner_timeout = 300

    client.load_system_host_keys()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    ssh_config = paramiko.SSHConfig()
    user_config_file = os.path.expanduser("~/.ssh/config")
    if os.path.exists(user_config_file):
        with open(user_config_file) as filename:
            ssh_config.parse(filename)
    else:
        print("SSH configuration file not found at %s" % user_config_file)

    cfg = {'hostname': hostname}

    user_config = ssh_config.lookup(cfg['hostname'])
    for ssh_cfg_name, paramiko_cfg_name in {'hostname': 'hostname',
                                            'user': 'username',
                                            'port': 'port'}.items():
        if ssh_cfg_name in user_config:
            cfg[paramiko_cfg_name] = user_config[ssh_cfg_name]

    if 'proxycommand' in user_config:
        cfg['sock'] = paramiko.ProxyCommand(user_config['proxycommand'])

    if 'identityfile' in user_config:
        cfg['key_filename'] = os.path.expanduser(user_config['identityfile'][0])
        if not os.path.exists(cfg['key_filename']):
            raise Exception("Specified IdentityFile " + cfg['key_filename'] + " for " +
                            hostname + " in ~/.ssh/config not existing anymore.")
    print("SSH configuration : %s" % cfg)
    client.connect(**cfg)
    return client


###############################################################################
#                                   scp_put                                   #
###############################################################################

def scp_put(hostname, src_path, dst_path):
    """ Copy from local to remote using SCP """
    ssh = _create_ssh_client(hostname)
    with SCPClient(ssh.get_transport()) as scp:
        scp.put(src_path, dst_path)


###############################################################################
#                                   scp_get                                   #
###############################################################################

def scp_get(hostname, src_path, dst_path):
    """ Copy from remote to local using SCP """
    ssh = _create_ssh_client(hostname)
    src_filename = os.path.basename(src_path)
    dst_filename = os.path.basename(dst_path)
    dst_dirname = os.path.dirname(dst_path)
    with Cd(dst_dirname):
        with SCPClient(ssh.get_transport()) as scpclient:
            scpclient.get(src_path)
        print("Renaming %s into %s" % (os.path.join(dst_dirname, src_filename),
                                       os.path.join(dst_dirname, dst_filename)))
        os.rename(os.path.join(dst_dirname, src_filename),
                  os.path.join(dst_dirname, dst_filename))
    if not os.path.exists(dst_path):
        raise Exception("Failed to scp from %s:%s to %s" % (hostname, src_path, dst_path))


###############################################################################
#                                 silentremove                                #
###############################################################################

def silentremove(filename):
    """Remove file quietly"""
    try:
        os.remove(filename)
    except (OSError, IOError) as error:
        if error.errno != errno.ENOENT:  # errno.ENOENT = no such file or directory
            raise  # re-raise exception if a different error occurred


###############################################################################
#                             silentremovedirs                                #
###############################################################################

def silentremovedirs(dirname):
    """Remove directory quietly"""
    shutil.rmtree(dirname, ignore_errors=True)


###############################################################################
#                                   sha256sum                                 #
###############################################################################

# Python program to find SHA256 hexadecimal hash string of a file
def sha256sum(filename):
    with open(filename, "rb") as f:
        filebyte = f.read()  # read entire file as bytes
        readable_hash = hashlib.sha256(filebyte).hexdigest()
        return readable_hash


###############################################################################
#                                  utf8_decorator                             #
###############################################################################

def utf8_decorator(func):
    """ decorator for encoding utf8 on all unicode arguments """

    def wrapper(*args, **kwargs):
        """ function wrapper """
        new_args = (x.encode('utf-8') if isinstance(x, string_types) else x for x in args)
        new_kwargs = {k: v.encode('utf-8') if isinstance(v, string_types) else v for k, v in kwargs.items()}
        return func(*new_args, **new_kwargs)

    return wrapper


###############################################################################
#                                  Get arguments                              #
###############################################################################

def get_arguments_ordered_dict(function):
    if sys.version_info.major >= 3:
        # noinspection PyUnresolvedReferences
        args = OrderedDict([(key, None) for key in inspect.getfullargspec(undecorated(function)).args[1:]])
    else:
        # noinspection PyDeprecation
        args = OrderedDict([(key, None) for key in inspect.getargspec(undecorated(function)).args[1:]])
    return args


###############################################################################
#                               Spill number fixer                            #
###############################################################################

def spill_number_fixer_passes_calculator(run_number):
    if run_number in range(58, 79):
        passes = {
            "1s": "FirstPassOnSteroids",
            "2s": "SecondPassOnSteroids",
            "3s": "ThirdPassOnSteroids",
            "1r": "FirstPassRegular",
            "2r": "SecondPassRegular",
            "3r": "ThirdPassRegular",
            "4r": "FourthPassRegular"
        }
    else:
        passes = {
            "1r": "FirstPassRegular",
            "2r": "SecondPassRegular",
            "3r": "ThirdPassRegular",
            "4r": "FourthPassRegular"
        }

    if run_number == 59:
        passes.pop("3s", None)
    elif run_number == 65:
        passes.pop("3s", None)
    elif run_number == 66:
        passes.pop("3s", None)

    return ','.join(list(passes.values()))


###############################################################################
#                               Parse run number                              #
###############################################################################

def parse_run_number(file_name, run_number=None):
    # type: (str, Optional[int]) -> int
    if run_number is not None and run_number > 0:
        this_run_number = run_number
    else:
        match = re.search(r'([\d]+)_ecal_dif_[\d]+_tree.root$', file_name)
        if match is not None:
            this_run_number = int(match.group(1))
        else:
            raise ValueError("Cannot detect run number in file name '{}'".format(file_name))
    return this_run_number


###############################################################################
#                               Parse input files                             #
###############################################################################

def parse_input_files(input_path, tree_friends=None, run_number=None):
    # type: (str, Optional[List[str]], Optional[int]) -> Tuple[int, Detectors]

    detectors = wagascianpy.plotting.detector.Detectors()
    detectors.disable_all_difs()

    if os.path.isfile(input_path):
        run_number = parse_run_number(file_name=input_path, run_number=run_number)
        dif_id = extract_dif_id(path=input_path)
        detectors.enable_dif(dif_id)
        detectors.get_dif(dif_id).add_tree(root_file=input_path, tree_name=extract_raw_tree_name(input_path),
                                           tree_friends=tree_friends)
    elif os.path.isdir(input_path):
        for file_name in os.listdir(input_path):
            if re.search(r'_ecal_dif_[\d]+_tree.root', file_name) is not None:
                run_number = parse_run_number(file_name=file_name, run_number=run_number)
                dif_id = extract_dif_id(path=file_name)
                detectors.enable_dif(dif_id)
                root_file = os.path.join(input_path, file_name)
                detectors.get_dif(dif_id).add_tree(root_file=root_file, tree_name=extract_raw_tree_name(root_file),
                                                   tree_friends=tree_friends)
    return run_number, detectors


###############################################################################
#                                Randomize                                    #
###############################################################################

def randomize(nentries, low, up):
    # type: (int, int, int) -> List[int]
    random_numbers = []
    if nentries > 0:
        for _ in range(nentries):
            random_numbers.append(random.randrange(low, up))
    return random_numbers



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio
import os

from six import string_types
import wagascianpy.utils.utils
import wagascianpy.database.db_record
from datetime import datetime


def _hst_file_name_parse(root, file_name, file_list, start_time=None, stop_time=None):
    file_name_no_ext, file_extension = os.path.splitext(file_name)
    if file_extension.strip('.') == "hst":
        if start_time is None or stop_time is None:
            file_list.append(os.path.join(root, file_name))
        else:
            time = datetime.strptime(file_name_no_ext, "%y%m%d")
            time = wagascianpy.database.db_record.DBRecord.add_timezone(time)
            if start_time < time < stop_time:
                file_list.append(os.path.join(root, file_name))


def mhistory2sqlite(input_path, output_folder=None, start_time=None, stop_time=None, recursive=False):
    # Check input_path argument
    if not isinstance(input_path, string_types):
        raise TypeError("Input folder must be a string")
    if not input_path:
        raise ValueError("Input folder cannot be empty")
    if not os.path.exists(input_path) or not os.access(input_path, os.R_OK):
        raise OSError("Input folder must exists and be readable : %s" % input_path)

    # Check output_folder argument
    if not output_folder:
        if os.path.isdir(input_path):
            output_folder = input_path
        else:
            output_folder = os.path.dirname(input_path)
    if not isinstance(output_folder, string_types):
        raise TypeError("Output folder must be a string")
    if os.path.exists(output_folder) and not os.access(output_folder, os.R_OK):
        raise OSError("Output folder must exists and be readable : %s" % output_folder)
    if not os.path.exists(output_folder):
        wagascianpy.utils.utils.mkdir_p(output_folder)

    # Check recursive argument
    if not isinstance(recursive, bool):
        raise TypeError("Recursive flag must be a boolean")

    # Check start time and stop time
    if not isinstance(start_time, string_types):
        raise TypeError("Start time must be a string")
    start_time = wagascianpy.database.db_record.DBRecord.str2datetime(start_time)
    if not isinstance(stop_time, string_types):
        raise TypeError("Stop time must be a string")
    stop_time = wagascianpy.database.db_record.DBRecord.str2datetime(stop_time)

    # Check that the mh2sql program exists
    mh2sql = wagascianpy.utils.utils.which(program="mh2sql")
    if not mh2sql:
        raise EnvironmentError("mh2sql program was not found")

    # List the input files
    input_files = []
    if os.path.isfile(input_path):
        input_files.append(input_path)
    if os.path.isdir(input_path):
        if recursive:
            for root, dirs, files in os.walk(input_path):
                for file_name in files:
                    _hst_file_name_parse(root=root, file_name=file_name, file_list=input_files, start_time=start_time,
                                         stop_time=stop_time)
        else:
            for file_name in os.listdir(input_path):
                _hst_file_name_parse(root=input_path, file_name=file_name, file_list=input_files, start_time=start_time,
                                     stop_time=stop_time)

    input_files = sorted(input_files)
    files_list = ' '.join(input_files)

    cmd = "{} --sqlite {} {}".format(mh2sql, output_folder, files_list)
    print(wagascianpy.utils.utils.run_cmd(cmd))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2019 Pintaudi Giorgio

"""Check the gain of the SPIROC2D ASIC (method 1)"""

# Python modules
import argparse
import json
import os
import textwrap

import ROOT
from bitarray import bitarray

# WAGASCI modules
from wagascianpy.analysis.analysis import WagasciAnalysis, ThreadWithReturnValue
from wagascianpy.utils.environment import WagasciEnvironment
from wagascianpy.utils.utils import limit_threads, join_threads

MAX_NB_THREAD_CHAINS = 8
PEU_LEVEL = 1
IDAC_FORM = "iDACOPTIMIZE"
PEU_FORM = "PEU%s"


# ================================================================ #
#                                                                  #
#                        Analysis functions                        #
#                                                                  #
# ================================================================ #

def _gain_decoder(process, run_root_dir, peu, threshold, acq_name, dif, n_chips, overwrite_flag):
    # type: (WagasciAnalysis, str, int, str, str, int, int, bool) -> ThreadWithReturnValue
    """ wgDecoder for the gain_run.py script """
    peu = int(peu)
    dif = int(dif)
    n_chips = int(n_chips)
    overwrite_flag = bool(overwrite_flag)
    input_raw_file = os.path.join(run_root_dir, IDAC_FORM, PEU_FORM % peu, threshold,
                                  "RawData", "{}_ecal_dif_{}.raw".format(acq_name, dif))
    output_dir = os.path.join(run_root_dir, IDAC_FORM, PEU_FORM % peu, threshold, "wgDecoder")
    compatibility_mode = False
    return process.spawn("decoder", input_raw_file, "", output_dir, overwrite_flag, compatibility_mode,
                         dif, n_chips)


def _gain_make_hist(process, run_root_dir, peu, threshold, acq_name, acq_config_xml, dif):
    # type: (WagasciAnalysis, str, int, str, str, str, int) -> ThreadWithReturnValue
    """ wgMakeHist for the gain_check.py script """
    dif = int(dif)
    input_tree_file = os.path.join(run_root_dir, IDAC_FORM, PEU_FORM % peu, threshold, "wgDecoder",
                                   "{}_ecal_dif_{}_tree.root".format(acq_name, dif))
    output_dir = os.path.join(run_root_dir, IDAC_FORM, PEU_FORM % peu, threshold, "wgMakeHist")
    flags = bitarray('0' * 9, endian='big')
    flags[0] = True  # overwrite
    flags[4] = True  # charge hit HG
    flags[7] = True  # Dark noise
    ul_flags = int(flags.to01(), 2)
    return process.make_hist(input_file=input_tree_file, config_file=acq_config_xml, output_dir=output_dir,
                             ul_flags=ul_flags, dif=dif)


def _gain_gain_check(process, run_root_dir, acq_config_xml, only_wallmrd, only_wagasci):
    # type: (WagasciAnalysis, str, str, bool, bool) -> ThreadWithReturnValue
    """ wgGainCalib for the gain_check.py script """
    only_wallmrd = bool(only_wallmrd)
    only_wagasci = bool(only_wagasci)
    output_dir = os.path.join(run_root_dir, "wgGainCheck")
    return process.gain_check(input_dir=run_root_dir, xml_config_file=acq_config_xml, output_img_dir=output_dir,
                              only_wallmrd=only_wallmrd, only_wagasci=only_wagasci)


def _gain_gain_tune(process, run_root_dir, acq_config_xml, only_wallmrd, only_wagasci):
    # type: (WagasciAnalysis, str, str, bool, bool) -> ThreadWithReturnValue
    """ wgGainCalib for the gain_check.py script """
    only_wallmrd = bool(only_wallmrd)
    only_wagasci = bool(only_wagasci)
    output_dir = os.path.join(run_root_dir, "wgGainTune")
    return process.gain_tune(input_dir=run_root_dir, xml_config_file=acq_config_xml, output_img_dir=output_dir,
                             only_wallmrd=only_wallmrd, only_wagasci=only_wagasci)


# ============================================================================ #
#                                                                              #
#                                gain_analysis                                 #
#                                                                              #
# ============================================================================ #

def gain_check_analysis(run, only_wallmrd=False, only_wagasci=False, decoder=True, makehist=True,
                        overwrite=False, single_threshold=False):
    """ Analyze gain check data """

    only_wallmrd = bool(only_wallmrd)
    only_wagasci = bool(only_wagasci)
    decoder = bool(decoder)
    makehist = bool(makehist)
    single_threshold = bool(single_threshold)

    # Environmental variables
    env = WagasciEnvironment()
    wagasci_lib = env['WAGASCI_LIBDIR']

    if not os.path.exists(run):
        run = os.path.join(env['WAGASCI_CALIBDATADIR'], run)

    run_name = os.path.basename(run)
    run_root_dir = run

    acq_config_path = os.path.join(run_root_dir, os.path.basename(env['WAGASCI_ACQCONFIGDIR']))
    acq_config_xml = os.path.join(acq_config_path, env['WAGASCI_ACQCONFIGXML'])
    if not os.path.exists(acq_config_xml):
        raise EnvironmentError("Acquisition configuration XML file not found : %s" % acq_config_xml)
    if single_threshold:
        threshold_dirs = ["SingleTh"]
    else:
        threshold_dirs = ["LowTh", "MiddleTh", "HighTh"]

    # =========================================================== #
    #                        ANALYZE DATA                         #
    # =========================================================== #

    process = WagasciAnalysis(wagasci_lib)
    topology_string, pointer = process.get_dif_topology(acq_config_xml)
    dif_topology = json.loads(topology_string)
    process.free_topology(pointer)
    process.enable_thread_safety()
    del process

    ###########################################################################
    #                                 decoder                                 #
    ###########################################################################

    # The decoder is completely multithread safe so we can spawn as many threads
    # as the amount of available memory allows

    if decoder:

        threads = []

        for threshold_dir in threshold_dirs:
            for dif in dif_topology:
                if only_wagasci and int(dif) < 4:
                    continue
                if only_wallmrd and int(dif) >= 4:
                    continue
                acq_name = "{}_{}_{}".format(run_name, PEU_FORM % PEU_LEVEL, threshold_dir)
                filepath = os.path.join(run_root_dir, IDAC_FORM, PEU_FORM % PEU_LEVEL, threshold_dir, "wgDecoder",
                                        "{}_ecal_dif_{}_tree.root".format(acq_name, dif))
                if not overwrite and os.path.exists(filepath):
                    tfile = ROOT.TFile(filepath, "READ")
                    if not tfile.IsZombie():
                        print("Skipping %s" % filepath)
                        continue
                    else:
                        print("Removing zombie file %s" % filepath)
                        os.remove(filepath)
                limit_threads(threads, MAX_NB_THREAD_CHAINS)
                process = WagasciAnalysis(wagasci_lib)
                n_chips = len(dif_topology[dif])
                threads.append(_gain_decoder(process=process, run_root_dir=run_root_dir, peu=PEU_LEVEL,
                                             threshold=threshold_dir, acq_name=acq_name, dif=dif, n_chips=n_chips,
                                             overwrite_flag=overwrite))
                del process
            # dif loop
        # threshold loop

        # Wait until all the threads have returned
        join_threads(threads)

    ###########################################################################
    #                                make_hist                                #
    ###########################################################################

    if makehist:

        for threshold_dir in threshold_dirs:
            for dif in dif_topology:
                if only_wagasci and int(dif) < 4:
                    continue
                if only_wallmrd and int(dif) >= 4:
                    continue
                acq_name = "{}_{}_{}".format(run_name, PEU_FORM % PEU_LEVEL, threshold_dir)
                filepath = os.path.join(run_root_dir, IDAC_FORM, PEU_FORM % PEU_LEVEL, threshold_dir, "wgMakeHist",
                                        "{}_ecal_dif_{}_hist.root".format(acq_name, dif))
                if not overwrite and os.path.exists(filepath):
                    tfile = ROOT.TFile(filepath, "READ")
                    if not tfile.IsZombie():
                        print("Skipping %s" % filepath)
                        continue
                    else:
                        print("Removing zombie file %s" % filepath)
                        os.remove(filepath)
                process = WagasciAnalysis(wagasci_lib)
                result = _gain_make_hist(process=process, run_root_dir=run_root_dir, peu=PEU_LEVEL,
                                         threshold=threshold_dir, acq_name=acq_name, acq_config_xml=acq_config_xml,
                                         dif=dif)
                if result != 0:
                    print("wgMakeHist returned error code %d" % result)
                    exit(result)
                del process
            # dif loop
        # threshold loop

    ###########################################################################
    #                             gain_check                                  #
    ###########################################################################

    process = WagasciAnalysis(wagasci_lib)
    threshold_dir = "SingleTh" if single_threshold else "MiddleTh"
    acq_name = "{}_{}_{}".format(run_name, PEU_FORM % PEU_LEVEL, threshold_dir)
    acq_config_xml = os.path.join(run_root_dir, IDAC_FORM, PEU_FORM % PEU_LEVEL, threshold_dir, "RawData",
                                  "{}.xml".format(acq_name))
    result = _gain_gain_check(process=process, run_root_dir=run_root_dir, acq_config_xml=acq_config_xml,
                              only_wallmrd=only_wallmrd, only_wagasci=only_wagasci)
    del process
    if result != 0:
        print("wgGainCheck returned error code %d" % result)
        exit(result)

    process = WagasciAnalysis(wagasci_lib)
    result = _gain_gain_tune(process=process, run_root_dir=run_root_dir, acq_config_xml=acq_config_xml,
                             only_wallmrd=only_wallmrd, only_wagasci=only_wagasci)
    del process
    if result != 0:
        print("wgGainTune returned error code %d" % result)
        exit(result)


if __name__ == "__main__":
    PARSER = argparse.ArgumentParser(usage='use "python %(prog)s --help" for more information',
                                     argument_default=None, description='Check and tune the gain calibration')

    PARSER.add_argument('-f', '--run', metavar='<run name or path>', type=str, required=True,
                        help='Name or path of the run to analyze. If only the run name is given the run must be inside '
                             'the ${WAGASCI_CALIBDATADIR} folder.')
    PARSER.add_argument('--only-wallmrd', dest='only_wallmrd', action='store_true', help="Only analyze WallMRD data")
    PARSER.add_argument('--only-wagasci', dest='only_wagasci', action='store_true', help="Only analyze WAGASCI data")
    PARSER.add_argument('-d', '--disable-decoder', dest='decoder', action='store_false',
                        required=False, default=True, help="Disable the wgDecoder.")
    PARSER.add_argument('-m', '--disable-makehist', dest='makehist', action='store_false',
                        required=False, default=True, help="Disable the wgMakeHist.")
    PARSER.add_argument('-r', '--overwrite', dest='overwrite', action='store_true',
                        required=False, default=False, help="Overwrite flag.")
    PARSER.add_argument('-s', '--single-threshold', dest='single_threshold', action='store_true',
                        required=False, default=False, help=textwrap.dedent('''
                        If not set it is assumed that the low threshold LowTh, high threshold HighTh and middle
                        threshold MiddleTh folders are present. If set only the SingleTh folder is analyzed.
                        '''))

    ARGS = PARSER.parse_args()

    RUN = ARGS.run
    DECODER = ARGS.decoder
    MAKEHIST = ARGS.makehist
    OVERWRITE = ARGS.overwrite
    ONLY_WAGASCI = ARGS.only_wagasci
    ONLY_WALLMRD = ARGS.only_wallmrd
    SINGLE_THRESHOLD = ARGS.single_threshold

    gain_check_analysis(run=RUN, only_wallmrd=ONLY_WALLMRD, only_wagasci=ONLY_WAGASCI,
                        decoder=DECODER, makehist=MAKEHIST, overwrite=OVERWRITE, single_threshold=SINGLE_THRESHOLD)



import unittest

# noinspection PyProtectedMember
import wagascianpy.viewer.parse_args
# noinspection PyProtectedMember
from wagascianpy.viewer.configuration import fill_configuration, Configuration, RepositoryType


class TestParseArgs(unittest.TestCase):

    def setUp(self):
        self.args = [
            '--batch-mode',
            '--wagasci-repository=buyaka1',
            '--simple-wagasci-repository',
            '--borg-wagasci-repository',
            '--update-wagasci-database',
            '--rebuild-wagasci-database',
            '--wagasci-database=buyaka2',
            '--wagasci-libdir=buyaka3',
            '--bsd-repository=buyaka4',
            '--bsd-database=buyaka5',
            '--bsd-download-location=buyaka6',
            '--update-bsd-database',
            '--rebuild-bsd-database',
            '--wagasci-download-location=buyaka7',
            '--wagasci-decoded-location=buyaka8',
            '--temperature-sqlite-database=buyaka9',
            '--history-location=buyaka10',
            '--t2krun=10',
            '--start-time=today',
            '--stop-time=tomorrow',
            '--start-run=1',
            '--stop-run=2',
            '--only-good-runs',
            '--include-overlapping',
            '--download',
            '--decoder',
            '--bcid-distribution',
            '--adc-distribution',
            '--spill-number-fixer',
            '--beam-summary-data',
            '--adc-calibration',
            '--temperature',
            '--all-analyzers',
            '--overwrite-flag'
        ]

        self.parsed_args = {
            'batch_mode': True,
            'wagasci_repository': 'buyaka1',
            'simple_wagasci_repository': True,
            'borg_wagasci_repository': True,
            'update_wagasci_database': True,
            'rebuild_wagasci_database': True,
            'wagasci_database': 'buyaka2',
            'wagasci_libdir': 'buyaka3',
            'bsd_repository': 'buyaka4',
            'bsd_database': 'buyaka5',
            'bsd_download_location': 'buyaka6',
            'update_bsd_database': True,
            'rebuild_bsd_database': True,
            'wagasci_download_location': 'buyaka7',
            'wagasci_decoded_location': 'buyaka8',
            'temperature_sqlite_database': 'buyaka9',
            'history_location': 'buyaka10',
            't2krun': 10,
            'start_time': 'today',
            'stop_time': 'tomorrow',
            'start_run': 1,
            'stop_run': 2,
            'only_good_runs': True,
            'include_overlapping': True,
            'download': True,
            'decoder': True,
            'bcid_distribution': True,
            'adc_distribution': True,
            'spill_number_fixer': True,
            'beam_summary_data': True,
            'adc_calibration': True,
            'temperature': True,
            'all_analyzers': True,
            'overwrite_flag': True
        }

    def test_do_not_crash(self):
        wagascianpy.viewer.parse_args.parse_args(self.args)

    def test_list_flattening(self):
        parsed_args = wagascianpy.viewer.parse_args.parse_args(self.args)
        for name, value in vars(parsed_args).items():
            self.assertNotIsInstance(value, list)

    def test_correct_parsing(self):
        parsed_args = wagascianpy.viewer.parse_args.parse_args(self.args)
        for name, value in self.parsed_args.items():
            self.assertEqual(value, getattr(parsed_args, name))

    def test_get_time_interval_configuration(self):
        self.parsed_args['batch_mode'] = True
        self.parsed_args['start_run'] = 0
        self.parsed_args['stop_run'] = 0
        self.parsed_args['start_time'] = 'today'
        self.parsed_args['stop_time'] = 'tomorrow'
        self.parsed_args['borg_wagasci_repository'] = False
        fill_configuration(self.parsed_args)
        self.assertEqual(self.parsed_args['start_time'], Configuration.run_select.start_time())
        self.assertEqual(self.parsed_args['stop_time'], Configuration.run_select.stop_time())
        self.assertEqual(self.parsed_args['start_run'], Configuration.run_select.start_run())
        self.assertEqual(self.parsed_args['stop_run'], Configuration.run_select.stop_run())
        self.assertTrue(Configuration.run_select.get_time_interval())
        self.assertFalse(Configuration.run_select.get_run_interval())
        self.assertFalse(Configuration.run_select.get_all())

    def test_get_run_interval_configuration(self):
        self.parsed_args['batch_mode'] = True
        self.parsed_args['start_run'] = 1
        self.parsed_args['stop_run'] = 10
        self.parsed_args['start_time'] = None
        self.parsed_args['stop_time'] = None
        self.parsed_args['borg_wagasci_repository'] = False
        fill_configuration(self.parsed_args)
        self.assertEqual(self.parsed_args['start_time'], Configuration.run_select.start_time())
        self.assertEqual(self.parsed_args['stop_time'], Configuration.run_select.stop_time())
        self.assertEqual(self.parsed_args['start_run'], Configuration.run_select.start_run())
        self.assertEqual(self.parsed_args['stop_run'], Configuration.run_select.stop_run())
        self.assertFalse(Configuration.run_select.get_time_interval())
        self.assertTrue(Configuration.run_select.get_run_interval())
        self.assertFalse(Configuration.run_select.get_all())

    def test_get_all_configuration(self):
        self.parsed_args['batch_mode'] = True
        self.parsed_args['start_run'] = 0
        self.parsed_args['stop_run'] = 0
        self.parsed_args['start_time'] = None
        self.parsed_args['stop_time'] = None
        self.parsed_args['borg_wagasci_repository'] = False
        fill_configuration(self.parsed_args)
        self.assertEqual(self.parsed_args['start_time'], Configuration.run_select.start_time())
        self.assertEqual(self.parsed_args['stop_time'], Configuration.run_select.stop_time())
        self.assertEqual(self.parsed_args['start_run'], Configuration.run_select.start_run())
        self.assertEqual(self.parsed_args['stop_run'], Configuration.run_select.stop_run())
        self.assertFalse(Configuration.run_select.get_time_interval())
        self.assertFalse(Configuration.run_select.get_run_interval())
        self.assertTrue(Configuration.run_select.get_all())

    def test_change_configuration(self):
        self.parsed_args['batch_mode'] = True
        self.parsed_args['borg_wagasci_repository'] = False
        self.parsed_args['simple_wagasci_repository'] = True
        fill_configuration(self.parsed_args)
        # WAGASCI database configuration
        self.assertEqual(self.parsed_args['wagasci_database'],
                         Configuration.wagascidb.wagasci_database())
        self.assertEqual(self.parsed_args['wagasci_repository'],
                         Configuration.wagascidb.wagasci_repository())
        self.assertEqual(self.parsed_args['wagasci_download_location'],
                         Configuration.wagascidb.wagasci_download_location())
        self.assertEqual(self.parsed_args['wagasci_decoded_location'],
                         Configuration.wagascidb.wagasci_decoded_location())
        self.assertEqual(RepositoryType.Simple,
                         Configuration.wagascidb.repository_type())
        # BSD database configuration
        self.assertEqual(self.parsed_args['bsd_repository'],
                         Configuration.bsddb.bsd_repository())
        self.assertEqual(self.parsed_args['bsd_database'],
                         Configuration.bsddb.bsd_database())
        self.assertEqual(self.parsed_args['bsd_download_location'],
                         Configuration.bsddb.bsd_download_location())
        # global configuration
        self.assertEqual(self.parsed_args['t2krun'],
                         Configuration.global_configuration.t2krun())
        self.assertEqual(self.parsed_args['history_location'],
                         Configuration.global_configuration.history_location())
        # temperature configuration
        self.assertEqual(self.parsed_args['temperature_sqlite_database'],
                         Configuration.temperature.temperature_sqlite_database())

    def test_repository_type(self):
        self.parsed_args['borg_wagasci_repository'] = True
        self.parsed_args['simple_wagasci_repository'] = False
        fill_configuration(self.parsed_args)
        self.assertEqual(RepositoryType.Borg,
                         Configuration.wagascidb.repository_type())


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio
#

from __future__ import print_function

import json
import os
import random
import re
import subprocess
import time

from typing import Type, List, Callable, Optional

from wagascianpy.utils.configuration import Configuration, RepositoryType
from wagascianpy.database.wagascidb import WagasciRunRecord
import wagascianpy.utils.utils
from wagascianpy.viewer.import_tkinter import import_filedialog, import_pygubu
import wagascianpy.viewer.utils
from wagascianpy.viewer.topology import Topology


def downloader(config,  # type: Type[Configuration]
               records,  # type: List[WagasciRunRecord]
               topology,  # type: Topology
               reporter,  # type: Callable
               overwrite=False,  # type: bool
               batch_mode=False  # type: bool
               ):
    """Download the selected runs listed in the run list frame"""

    filedialog = import_filedialog()
    if not batch_mode and not filedialog:
        raise ImportError("filedialog module could not be imported!")

    if records is None:
        raise ValueError('First get some runs using the "Get All" or "Get Interval" buttons')

    # Ask the user for download folder
    download_folder = config.wagascidb.wagasci_download_location()
    if not download_folder:
        if not batch_mode:
            download_folder = filedialog.askdirectory()
        else:
            download_folder = input('Download folder')

    # Create the download directory if necessary
    wagascianpy.utils.utils.mkdir_p(download_folder)

    # Check for write permissions
    if not os.access(download_folder, os.W_OK):
        raise OSError("folder where to restore backup not writable")

    if config.wagascidb.repository_type() == RepositoryType.Borg:
        _borg_downloader(records=records,
                         wagasci_repository=config.wagascidb.wagasci_repository(),
                         topology=topology,
                         reporter=reporter,
                         overwrite=overwrite,
                         download_folder=download_folder,
                         batch_mode=batch_mode)
    else:
        _simple_downloader(records=records,
                           wagasci_repository=config.wagascidb.wagasci_repository(),
                           topology=topology,
                           reporter=reporter,
                           download_folder=download_folder,
                           batch_mode=batch_mode)
    return download_folder


# noinspection PyUnresolvedReferences
def _borg_downloader(records,  # type: List[wagascianpy.database.wagascidb.WagasciRunRecord]
                     wagasci_repository,  # type: str
                     topology,  # type: Topology
                     reporter,  # type: Callable
                     overwrite=True,  # type: bool
                     download_folder=None,  # type: Optional[str]
                     batch_mode=False  # type: bool
                     ):
    # Move to download folder
    with wagascianpy.utils.utils.Cd(download_folder):
        # Create list of borg extract commands
        borg_cmd_list = []
        for record in records:
            borg_extract_header = wagascianpy.utils.utils.which("borg") + " extract --progress --log-json " \
                                  + "--exclude *.json --exclude *.py --exclude *AcqConfig* " \
                                  + wagasci_repository + "::" + record["name"] + " "
            random_raw_file = random.choice(list(record["raw_files"].values()))
            # Extract whole run folder if all detectors are enabled
            if not overwrite:
                run_root_dir = os.path.join(download_folder, os.path.dirname(random_raw_file).strip('/'))
                if os.path.exists(run_root_dir):
                    continue
            if topology.are_all_enabled() is True:
                borg_extract = borg_extract_header + os.path.dirname(random_raw_file).strip('/')
                borg_cmd_list.append(borg_extract)
                del borg_extract
            else:
                # Else extract only the selected sub-detectors
                borg_extract = "%s%s/%s.xml" \
                               % (borg_extract_header, os.path.dirname(random_raw_file).strip('/'), record["name"])
                borg_cmd_list.append(borg_extract)
                del borg_extract
                borg_extract = "%s%s/%s.log" \
                               % (borg_extract_header, os.path.dirname(random_raw_file).strip('/'), record["name"])
                borg_cmd_list.append(borg_extract)
                del borg_extract
                for dif_id in (str(dif.index) for dif in topology.get_enabled()):
                    borg_extract = borg_extract_header + record["raw_files"][dif_id].strip('/')
                    borg_cmd_list.append(borg_extract)
                    del borg_extract

        # Return if there is nothing to download
        if not borg_cmd_list:
            return download_folder

        # Create download progress window
        print("Starting the download ...")
        if not batch_mode:
            pygubu = import_pygubu()
            if not pygubu:
                raise ImportError("pygubu module could not be imported!")
            builder = pygubu.Builder()
            builder.add_from_file(wagascianpy.viewer.utils.gui_file_finder())
            download_window = builder.get_object('download_window')
            download_label_text = builder.get_variable('download_label_text')
            download_label_text.set("Starting the download ...")
            progress_bar = builder.get_object("progress_bar")
            download_window.update()

        # Print list of commands to file for future reference
        with open(download_folder + "/borg_commands.txt", "w+") as borg_cmd_file:
            borg_cmd_string = '\n'.join(borg_cmd_list)
            borg_cmd_file.write(borg_cmd_string)

        # Actual download
        lost_run = None
        for borg_cmd in borg_cmd_list:
            if lost_run is not None and lost_run in borg_cmd:
                continue
            # Spawn a thread with the download process
            progress = wagascianpy.utils.utils.run_borg_cmd(borg_cmd, check_progress=True)
            countdown = 86400  # one day
            # Monitor the thread for progress
            while True:
                time.sleep(1)
                line = progress.stdout.readline()
                if not line:
                    raise RuntimeError("No answer from borgbackup")
                json_out = json.loads(line.decode('utf-8'))
                # If an error occurs do not block the download of all other runs
                if 'levelname' in json_out and json_out['levelname'] in [u'WARNING', u'ERROR', u'CRITICAL']:
                    if reporter:
                        reporter(exc=None, val=json_out['message'], tb=None)
                    if not batch_mode:
                        download_label_text.set(str(json_out['message']))
                    match = re.search(r"Archive ([\S]+) does not exist", str(json_out['message']))
                    if match is not None and len(match.groups()) >= 1:
                        lost_run = match.group(1)
                    break
                # Print the message in the download window
                if 'message' in json_out:
                    if not batch_mode:
                        download_label_text.set(str(json_out['message']))
                        print(download_label_text.get())
                # Calculate progress
                current = float(json_out['current']) if 'current' in json_out else 0
                total = float(json_out['total']) if 'total' in json_out else 0
                percentage = 100 * current / total if total != 0 else 0
                # Exit loop if the download is finished
                if 'finished' in json_out and json_out['finished'] is True or countdown <= 0:
                    break
                if not batch_mode:
                    progress_bar['value'] = percentage
                countdown -= 1
                # Update the window every second
                download_window.update_idletasks()
        if not batch_mode:
            progress_bar['value'] = 100
            download_label_text.set("Download complete (you may close this window now)")
            del download_label_text
            del progress_bar


# noinspection PyUnboundLocalVariable, PyUnresolvedReferences
def _simple_downloader(records,  # type: List[WagasciRunRecord]
                       wagasci_repository,  # type: str
                       topology,  # type: Topology
                       reporter,  # type: Callable
                       download_folder=None,  # type: Optional[str]
                       batch_mode=False  # type: bool
                       ):
    if ':' in wagasci_repository:
        if len(wagasci_repository.split(':', 1)) != 2:
            raise ValueError("Invalid repository location : %s" % wagasci_repository)
        is_remote_repo = True
    else:
        is_remote_repo = False

    # Create download progress window
    print("Starting the download ...")
    if not batch_mode:
        pygubu = import_pygubu()
        if not pygubu:
            raise ImportError("pygubu module could not be imported!")
        builder = pygubu.Builder()
        builder.add_from_file(wagascianpy.viewer.utils.gui_file_finder())
        download_window = builder.get_object('download_window')
        download_label_text = builder.get_variable('download_label_text')
        download_label_text.set("Starting the download ...")
        progress_bar = builder.get_object("progress_bar")
        download_window.update()

    # Move to download folder
    if not is_remote_repo:
        if not os.path.isdir(wagasci_repository):
            raise EnvironmentError("Local WAGASCI repository not found at %s" % wagasci_repository)

    # List all sources to download
    download_list = []
    for record in records:
        if topology.are_all_enabled() is True:
            source = os.path.join(wagasci_repository, record["name"])
            destination = download_folder
            download_list.append((source, destination))
            print("Copying repository %s into location %s" % (source, destination))
        else:
            destination = os.path.join(download_folder, record["name"])
            source = os.path.join(wagasci_repository, record["name"], record["name"] + ".xml")
            download_list.append((source, destination))
            source = os.path.join(wagasci_repository, record["name"], record["name"] + ".log")
            download_list.append((source, destination))
            for dif in topology.get_enabled():
                source = os.path.join(wagasci_repository, record["name"],
                                      "{}_ecal_dif_{}.raw".format(record["name"], dif.index))
                download_list.append((source, destination))
                source = os.path.join(wagasci_repository, record["name"],
                                      "{}_ecal_dif_{}_tree.root".format(record["name"], dif.index))
                download_list.append((source, destination))
                print("Copying DIF %s into location %s" % (source, destination))

    # Actual download
    with wagascianpy.utils.utils.Cd(download_folder):
        for source, destination in download_list:
            wagascianpy.utils.utils.mkdir_p(destination)
            rsync = wagascianpy.utils.utils.which("rsync")
            if rsync is None:
                raise RuntimeError("rsync program not found")

            if not batch_mode:
                download_label_text.set("Copying {} into the local directory {}".format(source, destination))

            if is_remote_repo:
                cmd_ending = '-essh {} {}'.format(source, destination)
            else:
                cmd_ending = '{} {}'.format(source, destination)

            cmd = 'rsync -avz --progress ' + cmd_ending
            try:
                wagascianpy.utils.utils.mkdir_p(destination)
                proc = subprocess.Popen(cmd,
                                        shell=True,
                                        stdin=subprocess.PIPE,
                                        stdout=subprocess.PIPE,
                                        universal_newlines=True
                                        )
            except subprocess.CalledProcessError as exception:
                error_message = "Error while copying the repository {} into the local " \
                                "directory {} : {}".format(source, destination, str(exception))
                _report_error(error_message=error_message, reporter=reporter)

            for line in iter(proc.stdout.readline, ''):
                print(line, end='')
                if '%' in line:
                    progress = re.findall(r'(\d+)%', line)[0]
                    if not batch_mode:
                        progress_bar['value'] = progress
                        # Update the window
                        download_window.update_idletasks()
                elif 'total size is' in line:
                    break

        if not batch_mode:
            progress_bar['value'] = 100
            download_label_text.set("Download complete (you may close this window now)")
            download_window.update_idletasks()
            del download_label_text
            del progress_bar


def _report_error(error_message, reporter=None):
    if reporter:
        reporter(exc=None, val=error_message, tb=None)
    else:
        print(error_message)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

# Python modules
import argparse
import re

DIF_IFX = 0
CHIP_IDX = 1
CHAN_IDX = 2
POSX_IDX = 3
POSY_IDX = 4
POSZ_IDX = 5
VIEW_IDX = 6
PLANE_IDX = 7
SLOT_IDX = 8
PX_IDX = 9

CHANNEL_ID_TO_PX = {
    15: 1,
    11: 2,
    14: 3,
    10: 4,
    13: 5,
    9: 6,
    12: 7,
    8: 8,
    0: 9,
    1: 10,
    2: 11,
    3: 12,
    4: 13,
    5: 14,
    6: 15,
    7: 16,
    26: 17,
    23: 18,
    27: 19,
    22: 20,
    25: 21,
    21: 22,
    24: 23,
    20: 24,
    28: 25,
    19: 26,
    29: 27,
    18: 28,
    30: 29,
    17: 30,
    31: 31,
    16: 32
}

if __name__ == "__main__":
    PARSER = argparse.ArgumentParser(usage='use "python %(prog)s --help" for more information',
                                     argument_default=None, description='Insert PX information into mapping file')

    PARSER.add_argument('-f', '--mapping-file-input', metavar='<mapping file>', dest='mapping_file_input', type=str,
                        required=True, help='Path to the input WallMRD mapping file')

    PARSER.add_argument('-o', '--mapping-file-output', metavar='<mapping file>', dest='mapping_file_output', type=str,
                        required=True, help='Path to the output WallMRD mapping file')

    ARGS = PARSER.parse_args()

    with open(ARGS.mapping_file_input, "r") as mapping_file_input:
        with open(ARGS.mapping_file_output, "w") as mapping_file_output:
            repl_str = re.compile('^[-+]?\d*\.\d+|\d+$')
            input_lines = mapping_file_input.readlines()
            counter = 0
            for input_line in input_lines:
                counter += 1
                input_fields = []
                input_words = re.split(' |\t', input_line)
                for input_word in input_words:
                    match = re.search(repl_str, input_word)
                    if match:
                        input_fields.append(float(match.group()))
                if len(input_fields) == 10:
                    mapping_file_output.write(input_line)
                elif len(input_fields) == 9:
                    output_line = "{}\t{}\n".format(input_line.strip('\n'), CHANNEL_ID_TO_PX[input_fields[CHAN_IDX]])
                    mapping_file_output.write(output_line)
                else:
                    if input_line[0] == '#':
                        mapping_file_output.write(input_line)
                    else:
                        print("Line not recognized : %s" % input_line)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2019 Pintaudi Giorgio

""" Run script to measure and check the pedestal of the SPIROC2D ASIC """

# pylint: disable-msg=too-many-arguments
# pylint: disable-msg=too-many-function-args
# pylint: disable-msg=too-many-locals

# Python modules
import os
import time
import json
import argparse
from bitarray import bitarray

# WAGASCI modules
from wagascianpy.analysis import WagasciAnalysis
from wagascianpy.utils.environment import WagasciEnvironment
from wagascianpy.utils import limit_threads, join_threads

RANGE_IDAC = [1, 121, 241]
RANGE_PEU = [1, 2]  # threshold set at 0.5 and 1.5 p.e.
MAX_NB_THREAD_CHAINS = 8


# ================================================================ #
#                                                                  #
#                        Analysis functions                        #
#                                                                  #
# ================================================================ #

def _scurve_check_decoder(process, run_root_dir, acq_name, dif, n_chips):
    """ wgDecoder for the scurve_check_run.py script """
    dif = int(dif)
    n_chips = int(n_chips)
    input_raw_file = "%s/RawData/%s_ecal_dif_%d.raw" % (run_root_dir, acq_name, dif)
    output_dir = "%s/wgDecoder" % run_root_dir
    overwrite_flag = True
    compatibility_mode = False
    disable_calibration_variables = True
    return process.spawn("decoder", input_raw_file, "", output_dir,
                         overwrite_flag, compatibility_mode,
                         disable_calibration_variables, dif, n_chips)


def _scurve_check_make_hist(process, run_root_dir, acq_name, acq_config_xml, dif):
    """ wgMakeHist for the scurve_check_run.py script """
    dif = int(dif)
    input_tree_file = "%s/wgDecoder/%s_ecal_dif_%d_tree.root" \
                      % (run_root_dir, acq_name, dif)
    output_dir = "%s/wgMakeHist" % run_root_dir
    flags = bitarray('0' * 9, endian='big')
    flags[7] = True  # dark noise
    flags[0] = True  # overwrite
    ul_flags = int(flags.to01(), 2)
    return process.make_hist(input_tree_file, acq_config_xml, output_dir,
                             ul_flags, dif)


def _scurve_check_ana_hist(process, run_root_dir, acq_name, acq_config_xml, dif):
    """ wgAnaHist for the scurve_check_run.py script """
    dif = int(dif)
    input_hist_file = "%s/wgMakeHist/%s_ecal_dif_%d_hist.root" \
                      % (run_root_dir, acq_name, dif)
    output_dir = "%s/wgAnaHist/Xml/dif_%d" % (run_root_dir, dif)
    output_img_dir = "%s/wgAnaHist/Images/dif_%d" % (run_root_dir, dif)
    flags = bitarray('0' * 8, endian='big')
    flags[7] = True  # overwrite
    flags[6] = bool(acq_config_xml)
    flags[5] = False  # print
    flags[4] = True  # Dark noise
    ul_flags = int(flags.to01(), 2)
    return process.spawn("ana_hist", input_hist_file, acq_config_xml, output_dir,
                         output_img_dir, ul_flags, dif)
    # return process.ana_hist(input_hist_file, acq_config_xml, output_dir,
    #                         output_img_dir, ul_flags, dif)


def _scurve_check_ana_hist_summary(process, run_root_dir, dif):
    """ wgAnaHistSummary for the scurve_check_run.py script """
    dif = int(dif)
    input_dir = "%s/wgAnaHist/Xml/dif_%d" % (run_root_dir, dif)
    output_dir = "%s/wgAnaHistSummary/Xml/dif_%d" % (run_root_dir, dif)
    output_img_dir = "%s/wgAnaHistSummary/Images/dif_%d" % (run_root_dir, dif)
    flags = bitarray('0' * 8, endian='big')
    flags[7] = True  # overwrite
    flags[5] = True  # print
    flags[4] = True  # Dark noise
    ul_flags = int(flags.to01(), 2)
    return process.ana_hist_summary(input_dir, output_dir, output_img_dir, ul_flags)


# ============================================================================ #
#                                                                              #
#                                scurve_check_analysis                                 #
#                                                                              #
# ============================================================================ #

def scurve_check_analysis(run_name, only_wallmrd=False, only_wagasci=False,
                          decoder=True, makehist=True, anahist=True):
    """ Analyze pedestal calibration data """

    only_wallmrd = bool(only_wallmrd)
    only_wagasci = bool(only_wagasci)
    decoder = bool(decoder)
    makehist = bool(makehist)
    anahist = bool(anahist)

    # Environmental variables
    env = WagasciEnvironment()
    calibdata_dir = env['WAGASCI_CALIBDATADIR']
    wagasci_lib = env["WAGASCI_LIBDIR"]

    run_root_dir = calibdata_dir + "/" + run_name
    acq_config_xml = run_root_dir + "/AcqConfig/wagasci_config.xml"

    # =========================================================== #
    #                        ANALYZE DATA                         #
    # =========================================================== #

    process = WagasciAnalysis(wagasci_lib)
    if not os.path.exists(acq_config_xml):
        raise ValueError("Acquisition XML configuration file does not exists : " +
                         acq_config_xml)
    topology_string, pointer = process.get_dif_topology(acq_config_xml)
    dif_topology = json.loads(topology_string)
    process.free_topology(pointer)
    process.enable_thread_safety()
    del process

    #############################################################################
    #                                   decoder                                 #
    #############################################################################

    # The decoder is completely multithread safe so we can spawn as many threads
    # as the amount of available memory allows

    if decoder:

        threads = []

        for dif in dif_topology:
            if only_wagasci and int(dif) < 4:
                continue
            if only_wallmrd and int(dif) >= 4:
                continue
            limit_threads(threads, MAX_NB_THREAD_CHAINS)
            process = WagasciAnalysis(wagasci_lib)
            n_chips = len(dif_topology[dif])
            threads.append(_scurve_check_decoder(process, run_root_dir,
                                                 run_name, dif, n_chips))
            del process
        # dif loop

        # Wait until all the threads have returned
        join_threads(threads)
        time.sleep(5)

    #############################################################################
    #                                  make_hist                                #
    #############################################################################

    # wgMakeHist does not play nicely with threads. Because of the wacky ROOT
    # memory handling there are a lot of seg faults and deadlocks when running
    # in a multithread environment. Moreover the memory footprint of wgMakeHist
    # is quite big (~1 GB per thread), so only a few threads can run
    # simultaneuously anyway. So be warned.

    if makehist:

        actual_acq_config_xml = "%s/RawData/%s.xml" % (run_root_dir, run_name)
        for dif in dif_topology:
            if only_wagasci and int(dif) < 4:
                continue
            if only_wallmrd and int(dif) >= 4:
                continue
            process = WagasciAnalysis(wagasci_lib)
            result = _scurve_check_make_hist(process, run_root_dir, run_name,
                                             actual_acq_config_xml, dif)
            if result != 0:
                print("wgMakeHist returned error code %d" % result)
                exit(result)
            del process
        # dif loop

    #############################################################################
    #                                   ana_hist                                #
    #############################################################################

    # wgAnaHist is thread safe if there is no drawing (no print image to file)

    if anahist:

        threads = []

        actual_acq_config_xml = "%s/RawData/%s.xml" % (run_root_dir, run_name)
        for dif in dif_topology:
            if only_wagasci and int(dif) < 4:
                continue
            if only_wallmrd and int(dif) >= 4:
                continue
            limit_threads(threads, MAX_NB_THREAD_CHAINS)
            process = WagasciAnalysis(wagasci_lib)
            threads.append(_scurve_check_ana_hist(process, run_root_dir, run_name,
                                                  actual_acq_config_xml, dif))
            # result = _scurve_check_ana_hist(process, run_root_dir, input_dac,
            #                         peu, acq_name,
            #                         actual_acq_config_xml, dif)
            # if result != 0:
            #     print "wgAnaHist returned error code %d" %(result)
            #     exit(result)
            del process
        # dif loop

        # Wait until all the threads have returned
        join_threads(threads)
        time.sleep(5)

    #############################################################################
    #                             ana_hist_summary                              #
    #############################################################################

    # The wgAnaHistSummary code is not thread safe

    if anahist:

        for dif in dif_topology:
            if only_wagasci and int(dif) < 4:
                continue
            if only_wallmrd and int(dif) >= 4:
                continue
            process = WagasciAnalysis(wagasci_lib)
            result = _scurve_check_ana_hist_summary(process, run_root_dir, dif)
            if result != 0:
                print("wgAnaHistSummary returned error code %d" % result)
                exit(result)
            del process
        # dif loop


###############################################################################
#                                  arguments                                  #
###############################################################################

if __name__ == "__main__":
    PARSER = argparse.ArgumentParser(description='Analyze scurve calibration data')

    PARSER.add_argument('run_name', metavar='N', type=str, nargs='+',
                        help='Name of the run to analyze')
    PARSER.add_argument('--only-wallmrd', dest='only_wallmrd', action='store_true')
    PARSER.add_argument('--only-wagasci', dest='only_wagasci', action='store_true')
    PARSER.add_argument('--disable-decoder', dest='decoder', action='store_false')
    PARSER.add_argument('--disable-makehist', dest='makehist', action='store_false')
    PARSER.add_argument('--disable-anahist', dest='anahist', action='store_false')

    PARSER.set_defaults(ignore_wagasci=False)

    ARGS = PARSER.parse_args()
    RUN_NAME = ARGS.run_name[0]
    ONLY_WAGASCI = False
    ONLY_WAGASCI = ARGS.only_wagasci
    ONLY_WALLMRD = False
    ONLY_WALLMRD = ARGS.only_wallmrd
    DECODER = True
    DECODER = ARGS.decoder
    MAKEHIST = True
    MAKEHIST = ARGS.makehist
    ANAHIST = True
    ANAHIST = ARGS.anahist

    scurve_check_analysis(RUN_NAME, ONLY_WALLMRD, ONLY_WAGASCI,
                          DECODER, MAKEHIST, ANAHIST)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio
#

import unittest
from datetime import datetime
import pytz

from wagascianpy.database.db_record import DBRecord


class TestBsdDatabaseCreation(unittest.TestCase):

    def setUp(self):
        self.expected_dt_string = "2020/01/01 01:01:01"
        self.expected_dt_string_tz = "2020/01/01 01:01:01 JST"
        self.expected_dt = datetime(year=2020, month=1, day=1, hour=1, minute=1, second=1)
        jst = pytz.timezone(DBRecord.get_timezone_str())
        self.expected_dt = jst.localize(self.expected_dt)
        self.expected_timestamp = 1577808061.0

    def test_str2datetime(self):
        dt = DBRecord.str2datetime(self.expected_dt_string)
        self.assertEqual(self.expected_dt, dt)

    def test_tzstr2datetime(self):
        dt = DBRecord.str2datetime(self.expected_dt_string_tz)
        self.assertEqual(self.expected_dt, dt)

    def test_timestamp2datetime(self):
        dt = DBRecord.timestamp2datetime(self.expected_timestamp)
        self.assertEqual(self.expected_dt, dt)

    def test_str2timestamp(self):
        timestamp = DBRecord.str2timestamp(self.expected_dt_string)
        self.assertEqual(self.expected_timestamp, timestamp)

    def test_tzstr2timestamp(self):
        timestamp = DBRecord.str2timestamp(self.expected_dt_string_tz)
        self.assertEqual(self.expected_timestamp, timestamp)

    def test_datetime2timestamp(self):
        timestamp = DBRecord.datetime2timestamp(self.expected_dt)
        self.assertEqual(self.expected_timestamp, timestamp)

    def test_timestamp2str(self):
        dt_string_tz = DBRecord.timestamp2str(self.expected_timestamp)
        self.assertEqual(self.expected_dt_string_tz, dt_string_tz)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021 Pintaudi Giorgio

import json

from typing import Optional

# ROOT
try:
    # noinspection PyPackageRequirements
    import ROOT

    ROOT.PyConfig.IgnoreCommandLineOptions = True
except ImportError:
    ROOT = None

from wagascianpy.utils.utils import parse_input_files
import wagascianpy.database.wagascidb
import wagascianpy.utils.treenames as treenames

SANITY_CHECK_ERROR_CODE = 1002

_TREE_FRIENDS = [treenames.BSD_TREE_NAME, treenames.FSN_TREE_NAME, treenames.PEU_TREE_NAME,
                 treenames.TIME_NS_TREE_NAME, treenames.FLAGS_TREE_NAME]


def sanity_check(input_path, wagasci_database, run_number=None):
    # type: (str, str, Optional[int]) -> None
    """
    Check the sanity of the decoded ROOT files. First it is checked if all the files are present by comparing the
    file list with the topology string. Then it is checked if all the TTrees inside the ROOT files are present.
    Finally it is checked if the number of spills in the trees is consistent.
    :param input_path: path to the input file or folder
    :param wagasci_database: path to the WAGASCI run database
    :param run_number: run number (optional)
    :return: True if all the sanity checks pass, False if one sanity check fails. In case of failure the following
    sanity checks are skipped.
    """
    if ROOT is None:
        raise ImportError("No ROOT module was found!")

    try:
        run_number, detectors = parse_input_files(input_path=input_path, run_number=run_number,
                                                  tree_friends=_TREE_FRIENDS)
    except ValueError as error:
        raise RuntimeError('Run number = {}, Input path = "{}" : {}'.format(run_number, input_path, error))

    with wagascianpy.database.wagascidb.WagasciDataBase(db_location=wagasci_database) as db:
        run_records = db.get_run_interval(run_number_start=run_number, only_good=False)
        if len(run_records) != 1:
            raise RuntimeError("Run {} not found in database".format(run_number))
        run_record = run_records[0]
        try:
            topology = json.loads(run_record.topology)
        except ValueError as error:
            raise RuntimeError('Error when parsing topology string "{}": {}'.format(topology, error))
        for dif_id in topology.keys():
            dif_id = int(dif_id)
            if not detectors.has_dif(dif_id):
                raise RuntimeError('Run number = {}, Input path = "{}" : DIF {} not found'.format(
                    run_number, input_path, dif_id))
            dif = detectors.get_dif(dif_id)
            num_spills = dif.get_tree().GetEntries()
            for friend in _TREE_FRIENDS:
                if not dif.has_friend(friend):
                    raise RuntimeError('Run number = {}, Input path = "{}", DIF = {} : tree "{}" not found'.format(
                        run_number, input_path, dif_id, friend))
                tree = dif.get_friend(friend)
                if tree.GetEntries() != num_spills:
                    raise RuntimeError('Run number = {}, Input path = "{}", DIF = {}, tree  = "{}" : num entries {} is '
                                       'different from {}'.format(run_number, input_path, dif_id, friend,
                                                                  tree.GetEntries(), num_spills))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

from enum import IntEnum
from typing import List, Union

import inflection

"""
The purpose of this module is to keep track of which detectors of DIFs are enabled or disabled. Simple as that.

The tricky part is that sometimes you need to need to list all enabled DIFs and sometimes a list of all the enabled 
detectors (which contain the DIFs), so the Topology class should be able to be iterated by DIF or by detector.
"""


def snakecase(string):
    # type: (str) -> str
    return string.lower().replace(' ', '_')


class DetectorType(IntEnum):
    """
    Enum representing an arbitrary detector type (WAGASCI or WallMRD)
    """
    Wagasci = 1
    Wallmrd = 2


class DetectorIndex(IntEnum):
    """
    Enum representing an arbitrary detector index
    """
    WallMrdNorth = 0,
    WallMrdSouth = 1,
    WagasciUpstream = 2,
    WagasciDownstream = 3,

    @classmethod
    def get_name(cls, index):
        # type: (int) -> str
        names = {
            cls.WallMrdNorth.value: "WallMRD north",
            cls.WallMrdSouth.value: "WallMRD south",
            cls.WagasciUpstream.value: "WAGASCI upstream",
            cls.WagasciDownstream.value: "WAGASCI downstream"
        }
        return names[index]

    @classmethod
    def get_name_snakecase(cls, index):
        # type: (int) -> str
        return snakecase(cls.get_name(index))


class DifIndex(IntEnum):
    """
    Enum representing a DIF index (as found in the DIF topology string)
    """
    WallmrdNorthTop = 0,
    WallmrdNorthBottom = 1,
    WallmrdSouthTop = 2,
    WallmrdSouthBottom = 3,
    WagasciUpstreamTop = 4,
    WagasciUpstreamSide = 5,
    WagasciDownstreamTop = 6,
    WagasciDownstreamSide = 7

    @classmethod
    def get_name(cls, index):
        # type: (int) -> str
        names = {
            cls.WallmrdNorthTop.value: DetectorIndex.get_name(DetectorIndex.WallMrdNorth) + " top",
            cls.WallmrdSouthTop.value: DetectorIndex.get_name(DetectorIndex.WallMrdSouth) + " top",
            cls.WagasciUpstreamTop.value: DetectorIndex.get_name(DetectorIndex.WagasciUpstream) + " top",
            cls.WagasciDownstreamTop.value: DetectorIndex.get_name(DetectorIndex.WagasciDownstream) + " top",
            cls.WallmrdNorthBottom.value: DetectorIndex.get_name(DetectorIndex.WallMrdNorth) + " bottom",
            cls.WallmrdSouthBottom.value: DetectorIndex.get_name(DetectorIndex.WallMrdSouth) + " bottom",
            cls.WagasciUpstreamSide.value: DetectorIndex.get_name(DetectorIndex.WagasciUpstream) + " side",
            cls.WagasciDownstreamSide.value: DetectorIndex.get_name(DetectorIndex.WagasciDownstream) + " side"
        }
        return names[index]

    @classmethod
    def get_index(cls, name):
        # type: (str) -> int
        return getattr(cls, inflection.camelize(snakecase(name)))

    @classmethod
    def get_name_snakecase(cls, index):
        # type: (int) -> str
        return snakecase(cls.get_name(index))

    @classmethod
    def is_wallmrd(cls, index):
        # type: (int) -> bool
        """
        :param index: DIF index
        :return: true if the DIF belongs to a WallMRD
        """
        return index in [int(cls.WallmrdNorthTop), int(cls.WallmrdNorthBottom),
                         int(cls.WallmrdSouthTop), int(cls.WallmrdSouthBottom)]

    @classmethod
    def is_wallmrd_north(cls, index):
        # type: (int) -> bool
        """
        :param index: DIF index
        :return: true if the DIF belongs to the north WallMRD
        """
        return index in [int(cls.WallmrdNorthTop), int(cls.WallmrdNorthBottom)]

    @classmethod
    def is_wallmrd_south(cls, index):
        # type: (int) -> bool
        """
        :param index: DIF index
        :return: true if the DIF belongs to the south WallMRD
        """
        return index in [int(cls.WallmrdSouthTop), int(cls.WallmrdSouthBottom)]

    @classmethod
    def is_wagasci(cls, index):
        # type: (int) -> bool
        """
        :param index: DIF index
        :return: true if the DIF belongs to a WallMRD
        """
        return index in [int(cls.WagasciUpstreamTop), int(cls.WagasciUpstreamSide),
                         int(cls.WagasciDownstreamTop), int(cls.WagasciDownstreamSide)]

    @classmethod
    def is_wagasci_upstream(cls, index):
        # type: (int) -> bool
        """
        :param index: DIF index
        :return: true if the DIF belongs to the north WallMRD
        """
        return index in [int(cls.WagasciUpstreamTop), int(cls.WagasciUpstreamSide)]

    @classmethod
    def is_wagasci_downstream(cls, index):
        # type: (int) -> bool
        """
        :param index: DIF index
        :return: true if the DIF belongs to the south WallMRD
        """
        return index in [int(cls.WagasciDownstreamTop), int(cls.WagasciDownstreamSide)]


class _State(object):
    """
    State of a sub-detector or of a DIF
    """

    def __init__(self, did, enabled, is_dif):
        # type: (Union[DifIndex, DetectorIndex], bool, bool) -> None
        """
        :param name: name of the sub-detector or DIF
        :param enabled: True if you want to enable it
        :param is_dif: True if is a DIF, False if it is a whole sub-detector
        """
        self.name = did.get_name(did.value)
        self._enabled = enabled
        self._is_dif = is_dif

    def is_enabled(self):
        # type: (...) -> bool
        """
        :return: True if is enabled
        """
        return self._enabled

    def enable(self):
        # type: (...) -> None
        """
        Enable the sub-detector of DIF
        :return: None
        """
        self._enabled = True

    def disable(self):
        # type: (...) -> None
        """
        Disable the sub-detector of DIF
        :return: None
        """
        self._enabled = False

    def is_dif(self):
        # type: (...) -> bool
        """
        :return: True if this object represents a DIF, False if it represents a whole sub-detector
        """
        return self._is_dif


class Topology(object):
    """List of _enabled and disabled detectors. If the flag '_iterate_by_dif' is disabled (default) the iterator
       iterates through the subdetectors. If the flag is _enabled the iterator iterates throught each DIF of the
       subdetectors."""

    def __init__(self, iterate_by_dif=False):
        # type: (bool) -> None
        """
        :param iterate_by_dif: True if you want the class iterator to iterate over every DIF, False if you want it to
        iterate only over the sub-detectors.
        """
        self._iterate_by_dif = iterate_by_dif
        if self._iterate_by_dif:
            self.wallmrd_north_top = _State(DifIndex.WallmrdNorthTop, True, True)
            self.wallmrd_north_bottom = _State(DifIndex.WallmrdNorthBottom, True, True)
            self.wallmrd_south_top = _State(DifIndex.WallmrdSouthTop, True, True)
            self.wallmrd_south_bottom = _State(DifIndex.WallmrdSouthBottom, True, True)
            self.wagasci_upstream_top = _State(DifIndex.WagasciUpstreamTop, True, True)
            self.wagasci_upstream_side = _State(DifIndex.WagasciUpstreamSide, True, True)
            self.wagasci_downstream_top = _State(DifIndex.WagasciDownstreamTop, True, True)
            self.wagasci_downstream_side = _State(DifIndex.WagasciDownstreamSide, True, True)
        else:
            self.wallmrd_north = _State(DetectorIndex.WallMrdNorth, True, False)
            self.wallmrd_south = _State(DetectorIndex.WallMrdSouth, True, False)
            self.wagasci_upstream = _State(DetectorIndex.WagasciUpstream, True, False)
            self.wagasci_downstream = _State(DetectorIndex.WagasciDownstream, True, False)

    @property
    def iterate_by_dif(self):
        """
        :return: True if the class iterator is iterating over all the DIFs, False if it is iterating only over the
        whole subdetectors.
        """
        return self._iterate_by_dif

    def import_topology(self, topology):
        # type: (Topology) -> None
        """
        Copy the topology object, while preserving its own iteration mode. The iteration mode of the other Topology
        object might be different.
        :param topology: other Topology object
        :return: None
        """
        if self.iterate_by_dif and not topology.iterate_by_dif:
            if topology.wagasci_upstream.is_enabled():
                self.wagasci_upstream_top.enable()
                self.wagasci_upstream_side.enable()
            else:
                self.wagasci_upstream_top.disable()
                self.wagasci_upstream_side.disable()
            if topology.wagasci_downstream.is_enabled():
                self.wagasci_downstream_top.enable()
                self.wagasci_downstream_side.enable()
            else:
                self.wagasci_downstream_top.disable()
                self.wagasci_downstream_side.disable()
            if topology.wallmrd_north.is_enabled():
                self.wallmrd_north_top.enable()
                self.wallmrd_north_bottom.enable()
            else:
                self.wallmrd_north_top.disable()
                self.wallmrd_north_bottom.disable()
            if topology.wallmrd_south.is_enabled():
                self.wallmrd_south_top.enable()
                self.wallmrd_south_bottom.enable()
            else:
                self.wallmrd_south_top.disable()
                self.wallmrd_south_bottom.disable()
        elif not self.iterate_by_dif and topology.iterate_by_dif:
            if topology.wagasci_upstream_top.is_enabled() or topology.wagasci_upstream_side.is_enabled():
                self.wagasci_upstream.enable()
            else:
                self.wagasci_upstream.disable()
            if topology.wagasci_downstream_top.is_enabled() or topology.wagasci_downstream_side.is_enabled():
                self.wagasci_downstream.enable()
            else:
                self.wagasci_downstream.disable()
            if topology.wallmrd_north_top.is_enabled() or topology.wallmrd_north_bottom.is_enabled():
                self.wallmrd_north.enable()
            else:
                self.wallmrd_north.disable()
            if topology.wallmrd_south_top.is_enabled() or topology.wallmrd_south_bottom.is_enabled():
                self.wallmrd_south.enable()
            else:
                self.wallmrd_south.disable()
        else:
            for key, value in vars(self).items():
                setattr(self, key, getattr(topology, key))

    def are_all_enabled(self):
        # type: (...) -> bool
        """
        :return: True if all DIFs are enabled
        """
        for value in vars(self).values():
            if isinstance(value, _State) and value.is_enabled() is False:
                return False
        return True

    def disable_all_but(self, one):
        # type: (str) -> None
        """
        Disable all the subdetectors or DIFs but the one with the specified name
        :param one: subdetector or DIF not to disable
        :return: None
        """
        one = snakecase(one)
        for d in vars(self).values():
            if isinstance(d, _State):
                if snakecase(d.name) == one:
                    d.enable()
                else:
                    d.disable()

    def enable(self, one):
        # type: (Union[int, str]) -> None
        """
        Enable a subdetector or DIF
        :param one: subdetector or DIF to enable
        :return: None
        """
        for d in vars(self).values():
            if isinstance(d, _State):
                if d.is_dif() == one:
                    d.enable()

    def how_many_enabled(self):
        # type: (...) -> int
        """
        :return: how many subdetectors or DIFs are enabled
        """
        return len(self.get_enabled())

    def disable_all(self):
        # type: (...) -> None
        """
        Disable everything
        :return: None
        """
        for d in vars(self).values():
            if isinstance(d, _State):
                d.disable()

    def get_all(self):
        # type: (...) -> List[_State]
        """
        :return: list of all the subdetectors or DIFs states (as objects of class _State)
        """
        if self._iterate_by_dif:
            return [value for value in vars(self).values() if isinstance(value, _State) and value.is_dif() is True]
        else:
            return [value for value in vars(self).values() if isinstance(value, _State) and value.is_dif() is False]

    def get_enabled(self):
        # type: (...) -> List[_State]
        """
        :return: list of enabled subdetectors or DIFs states (as objects of class _State)
        """
        if self._iterate_by_dif:
            return [value for value in vars(self).values()
                    if isinstance(value, _State) and value.is_dif() is True and value.is_enabled() is True]
        else:
            return [value for value in vars(self).values()
                    if isinstance(value, _State) and value.is_dif() is False and value.is_enabled() is True]

    def get_disabled(self):
        # type: (...) -> List[_State]
        """
        :return: list of disabled subdetectors or DIFs states (as objects of class _State)
        """
        if self._iterate_by_dif:
            return [value for value in vars(self).values()
                    if isinstance(value, _State) and value.is_dif() is True and value.is_enabled() is False]
        else:
            return [value for value in vars(self).values()
                    if isinstance(value, _State) and value.is_dif() is False and value.is_enabled() is False]

    def __iter__(self):
        # type: (...) -> TopologyIterator
        """
        Returns a TopologyIterator object (an iterator over the Topology class). The iterator iterates over the DIFs
        or over the sub-detectors depending on the iterator mode set through the iterate_by_dif constructor argument.
        :return: TopologyIterator object
        """
        return TopologyIterator(self)


class TopologyIterator(object):
    """
    Iterator over the Topology class
    """

    def __init__(self, topology):
        # type: (Topology) -> None
        """
        :param topology: Topology object which to iterate over
        """
        # Difs object reference
        self._topology = topology
        # member variable to keep track of current index
        self._index = 0

    def __next__(self):
        # type: (...) -> _State
        """
        :return: next DIF or subdetector
        """
        if self._topology.iterate_by_dif:
            if self._index == int(DifIndex.WallmrdNorthTop):
                result = self._topology.wallmrd_north_top
            elif self._index == int(DifIndex.WallmrdNorthBottom):
                result = self._topology.wallmrd_north_bottom
            elif self._index == int(DifIndex.WallmrdSouthTop):
                result = self._topology.wallmrd_south_top
            elif self._index == int(DifIndex.WallmrdSouthBottom):
                result = self._topology.wallmrd_south_bottom
            elif self._index == int(DifIndex.WagasciUpstreamTop):
                result = self._topology.wagasci_upstream_top
            elif self._index == int(DifIndex.WagasciUpstreamSide):
                result = self._topology.wagasci_upstream_side
            elif self._index == int(DifIndex.WagasciDownstreamTop):
                result = self._topology.wagasci_downstream_top
            elif self._index == int(DifIndex.WagasciDownstreamSide):
                result = self._topology.wagasci_downstream_side
            else:
                raise StopIteration
        else:
            if self._index == DetectorIndex.WallMrdNorth:
                result = self._topology.wallmrd_north
            elif self._index == DetectorIndex.WallMrdSouth:
                result = self._topology.wallmrd_south
            elif self._index == DetectorIndex.WagasciUpstream:
                result = self._topology.wagasci_upstream
            elif self._index == DetectorIndex.WagasciDownstream:
                result = self._topology.wagasci_downstream
            else:
                raise StopIteration
        self._index += 1
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2019 Pintaudi Giorgio

"""WAGASCI ANPAN for Python"""

import glob

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

extras = {
    'with_borg': ['borgbackup>=1.1.11']
}

setup(
    name='wagascianpy',
    version='0.3.9',
    author='Pintaudi Giorgio',
    author_email='giorgio.pimpa@gmail.com',
    include_package_data=True,
    zip_safe=True,
    keywords="WAGASCI ANPAN for Python",
    description='WAGASCI ANPAN for Python',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://git.t2k.org/wagasci_babymind/WagasciAnpy',
    packages=find_packages(),
    scripts=glob.glob('scripts/calibration/*.py') +
            glob.glob('scripts/data_quality/*.py') +
            glob.glob('scripts/mapping/*.py') +
            glob.glob('scripts/plotting/*.py') +
            glob.glob('scripts/viewer/*.py'),
    license='GPLv3',
    extras_require=extras,
    package_data={
        'wagascianpy': ['viewer/gui.ui', '*.rst', '**/*.rst'],
    },
    data_files=[
        ('wagasci/databases', glob.glob('../WagasciAnpy/databases/*')),
        ('wagasci/docs/images', glob.glob('../WagasciAnpy/docs/images/*'))
    ],
    install_requires=[
        'setuptools >= 72.1.0',
        'bitarray >= 0.8.1',
        'six >= 1.16.0',
        'paramiko >= 3.3.2',
        'scp >= 0.7.1',
        'pytz >= 2024.2',
        'tinydb >= 3.15.2',
        'recordclass >= 0.13.2',
        'pyfakefs >= 3.7.2',
        'numpy >= 1.26.4',
        'undecorated >= 0.3.0',
        'dependency_injector >= 3.15.6',
        'inflection >= 0.3.1',
        'pygubu >= 0.35.5',
        'typing >= 3.5.2.2',
        'mock >= 1.0.15.6'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: X11 Applications',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: Unix',
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2019 Pintaudi Giorgio

"""Script to calibration the gain of the high gain preamp the SPIROC2D ASIC

"""

# Python modules
import argparse
import json
import os
import re
from typing import Dict

import ROOT
from bitarray import bitarray

# WAGASCI modules
from wagascianpy.analysis.analysis import WagasciAnalysis, ThreadWithReturnValue
from wagascianpy.utils.environment import WagasciEnvironment
from wagascianpy.utils.utils import limit_threads, join_threads

PEU_LEVEL = 1
MAX_NB_THREAD_CHAINS = 8
IDAC_FORM = "iDAC%s"
PEU_FORM = "PEU%s"


# ================================================================ #
#                                                                  #
#                        Analysis functions                        #
#                                                                  #
# ================================================================ #

def _gain_decoder(process, run_root_dir, input_dac, peu, threshold_dir, acq_name, dif, n_chips, overwrite_flag):
    # type: (WagasciAnalysis, str, int, int, str, str, int, int, bool) -> ThreadWithReturnValue
    """ wgDecoder for the gain_run.py script """
    peu = int(peu)
    dif = int(dif)
    n_chips = int(n_chips)
    overwrite_flag = bool(overwrite_flag)
    input_raw_file = os.path.join(run_root_dir, IDAC_FORM % input_dac, PEU_FORM % peu, threshold_dir,
                                  "RawData", "%s_ecal_dif_%s.raw" % (acq_name, dif))
    output_dir = os.path.join(run_root_dir, IDAC_FORM % input_dac, PEU_FORM % peu, threshold_dir, "wgDecoder")
    compatibility_mode = False
    return process.spawn("decoder", input_raw_file, "", output_dir, overwrite_flag, compatibility_mode, dif, n_chips)


def _gain_make_hist(process, run_root_dir, input_dac, peu, threshold_dir, acq_name, acq_config_xml, dif):
    # type: (WagasciAnalysis, str, int, int, str, str, str, int) -> ThreadWithReturnValue
    """ wgMakeHist for the gain_run.py script """
    peu = int(peu)
    dif = int(dif)
    input_tree_file = os.path.join(run_root_dir, IDAC_FORM % input_dac, PEU_FORM % peu, threshold_dir,
                                   "wgDecoder", "%s_ecal_dif_%s_tree.root" % (acq_name, dif))
    output_dir = os.path.join(run_root_dir, IDAC_FORM % input_dac, PEU_FORM % peu, threshold_dir, "wgMakeHist")
    flags = bitarray('0' * 9, endian='big')
    flags[7] = True  # dark noise
    flags[4] = True  # charge hit HG
    flags[0] = True  # overwrite
    ul_flags = int(flags.to01(), 2)
    return process.make_hist(input_file=input_tree_file, config_file=acq_config_xml, output_dir=output_dir,
                             ul_flags=ul_flags, dif=dif)


def _gain_gain_calib(process, run_root_dir):
    # type: (WagasciAnalysis, str) -> ThreadWithReturnValue
    """ wgGainCalib for the gain_run.py script """
    output_xml_dir = os.path.join(run_root_dir, "wgGainCalib", "Xml")
    output_img_dir = os.path.join(run_root_dir, "wgGainCalib", "Images")
    return process.gain_calib(input_dir=run_root_dir, output_xml_dir=output_xml_dir, output_img_dir=output_img_dir,
                              print_flag=False)


def _get_idac_dirs(run_root_dir):
    # type: (str) -> Dict[int, str]
    dirs = {}
    for name in [name for name in os.listdir(run_root_dir)
                 if os.path.isdir(os.path.join(run_root_dir, name))]:
        idac = re.findall(r'\d+', name)
        if len(idac) == 1:
            dirs[int(idac[0])] = name
    return dirs


# ============================================================================ #
#                                                                              #
#                                gain_analysis                                 #
#                                                                              #
# ============================================================================ #

def gain_analysis(run, decoder=True, makehist=True, overwrite=False, single_threshold=False):
    # type: (str, bool, bool, bool, bool) -> None
    """ Analyze gain calibration data """

    decoder = bool(decoder)
    makehist = bool(makehist)
    single_threshold = bool(single_threshold)

    # Environmental variables
    env = WagasciEnvironment()
    wagasci_lib = env['WAGASCI_LIBDIR']

    if not os.path.exists(run):
        run = os.path.join(env['WAGASCI_CALIBDATADIR'], run)

    run_name = os.path.basename(run)
    run_root_dir = run

    acq_config_path = os.path.join(run_root_dir, os.path.basename(env['WAGASCI_ACQCONFIGDIR']))
    acq_config_xml = os.path.join(acq_config_path, env['WAGASCI_ACQCONFIGXML'])
    if not os.path.exists(acq_config_xml):
        raise EnvironmentError("Acquisition configuration XML file not found : %s" % acq_config_xml)

    idac_dirs = _get_idac_dirs(run_root_dir)

    if single_threshold:
        threshold_dirs = ["SingleTh"]
    else:
        threshold_dirs = ["LowTh", "MiddleTh", "HighTh"]

    # =========================================================== #
    #                        ANALYZE DATA                         #
    # =========================================================== #

    process = WagasciAnalysis(wagasci_lib)
    topology_string, pointer = process.get_dif_topology(acq_config_xml)
    dif_topology = json.loads(topology_string)
    process.free_topology(pointer)
    process.enable_thread_safety()
    del process

    ###########################################################################
    #                                 decoder                                 #
    ###########################################################################

    # The decoder is completely multithread safe so we can spawn as many threads
    # as the amount of available memory allows

    if decoder:

        threads = []

        for input_dac, idac_dir in idac_dirs.items():
            for threshold_dir in threshold_dirs:
                acq_name = "%s_iDAC%s_PEU%s_%s" % (run_name, input_dac, PEU_LEVEL, threshold_dir)
                for dif in dif_topology:
                    filepath = os.path.join(run_root_dir, idac_dir, PEU_FORM % PEU_LEVEL, threshold_dir,
                                            "wgDecoder", "%s_ecal_dif_%s_tree.root" % (acq_name, dif))
                    if not overwrite and os.path.exists(filepath):
                        tfile = ROOT.TFile(filepath, "READ")
                        if not tfile.IsZombie():
                            print("Skipping %s" % filepath)
                            continue
                        else:
                            print("Removing zombie file %s" % filepath)
                            os.remove(filepath)
                    limit_threads(threads, MAX_NB_THREAD_CHAINS)
                    process = WagasciAnalysis(wagasci_lib)
                    n_chips = len(dif_topology[dif])
                    threads.append(_gain_decoder(process=process, run_root_dir=run_root_dir, input_dac=input_dac,
                                                 peu=PEU_LEVEL, threshold_dir=threshold_dir, acq_name=acq_name,
                                                 dif=dif, n_chips=n_chips, overwrite_flag=overwrite))
                    del process
                # dif loop
            # threshold loop
        # input_dac loop

        # Wait until all the threads have returned
        join_threads(threads)

    ###########################################################################
    #                                make_hist                                #
    ###########################################################################

    if makehist:

        for input_dac in idac_dirs:
            for threshold_dir in threshold_dirs:
                acq_name = "%s_iDAC%s_PEU%s_%s" % (run_name, input_dac, PEU_LEVEL, threshold_dir)
                actual_acq_config_xml = os.path.join(run_root_dir, IDAC_FORM % input_dac, PEU_FORM % PEU_LEVEL,
                                                     threshold_dir, "RawData", "%s.xml" % acq_name)
                for dif in dif_topology:
                    filepath = os.path.join(run_root_dir, IDAC_FORM % input_dac, PEU_FORM % PEU_LEVEL,
                                            threshold_dir, "wgMakeHist", "%s_ecal_dif_%s_hist.root" % (acq_name, dif))
                    if not overwrite and os.path.exists(filepath):
                        tfile = ROOT.TFile(filepath, "READ")
                        if not tfile.IsZombie():
                            print("Skipping %s" % filepath)
                            continue
                        else:
                            print("Removing zombie file %s" % filepath)
                            os.remove(filepath)
                    process = WagasciAnalysis(wagasci_lib)
                    result = _gain_make_hist(process=process, run_root_dir=run_root_dir, input_dac=input_dac,
                                             peu=PEU_LEVEL, threshold_dir=threshold_dir, acq_name=acq_name,
                                             acq_config_xml=actual_acq_config_xml, dif=dif)
                    if result != 0:
                        print("wgMakeHist returned error code %s" % result)
                        exit(result)
                    del process
                # dif loop
            # threshold loop
        # input_dac loop

    ###########################################################################
    #                             gain_calib                                 #
    ###########################################################################

    process = WagasciAnalysis(wagasci_lib)
    result = _gain_gain_calib(process=process, run_root_dir=run_root_dir)
    del process
    if result != 0:
        print("wgGainCalib returned error code %s" % result)
        exit(result)


###############################################################################
#                                     MAIN                                    #
###############################################################################

if __name__ == "__main__":
    PARSER = argparse.ArgumentParser(usage='use "python %(prog)s --help" for more information',
                                     argument_default=None, description='Analyze the gain calibration data')

    PARSER.add_argument('-f', '--run', metavar='<run name or path>', type=str, required=True,
                        help='Name or path of the run to analyze. If only the run name is given the run must be inside '
                             'the ${WAGASCI_CALIBDATADIR} folder.')
    PARSER.add_argument('-d', '--disable-decoder', dest='decoder', action='store_false',
                        required=False, default=True, help="Disable the wgDecoder.")
    PARSER.add_argument('-m', '--disable-makehist', dest='makehist', action='store_false',
                        required=False, default=True, help="Disable the wgMakeHist.")
    PARSER.add_argument('-r', '--overwrite', dest='overwrite', action='store_true',
                        required=False, default=False, help="Overwrite flag.")

    ARGS = PARSER.parse_args()

    RUN = ARGS.run
    DECODER = ARGS.decoder
    MAKEHIST = ARGS.makehist
    OVERWRITE = ARGS.overwrite

    gain_analysis(run=RUN, decoder=DECODER, makehist=MAKEHIST, overwrite=OVERWRITE)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

# Python modules
import argparse
import re

DIF_IFX = 0
CHIP_IDX = 1
CHAN_IDX = 2
POSX_IDX = 3
POSY_IDX = 4
POSZ_IDX = 5
VIEW_IDX = 6
PLANE_IDX = 7
SLOT_IDX = 8

if __name__ == "__main__":
    PARSER = argparse.ArgumentParser(usage='use "python %(prog)s --help" for more information',
                                     argument_default=None, description='Check sanity of WallMRD mapping file')

    PARSER.add_argument('-f', '--mapping-file', metavar='<mapping file>', dest='mapping_file', type=str,
                        required=True, help='Path to the WallMRD mapping file')

    ARGS = PARSER.parse_args()

    with open(ARGS.mapping_file, "r") as mapping_file:
        repl_str = re.compile('^[-+]?\d*\.\d+|\d+$')
        lines = mapping_file.readlines()
        counter = 0
        for line in lines:
            counter += 1
            output = []
            words = re.split(' |\t', line)
            for word in words:
                match = re.search(repl_str, word)
                if match:
                    output.append(float(match.group()))
            if len(output) == 9:
                if (output[POSZ_IDX] == 100 and output[SLOT_IDX] != 0) or \
                        (output[POSZ_IDX] == 300 and output[SLOT_IDX] != 1) or \
                        (output[POSZ_IDX] == 500 and output[SLOT_IDX] != 2) or \
                        (output[POSZ_IDX] == 700 and output[SLOT_IDX] != 3) or \
                        (output[POSZ_IDX] == 900 and output[SLOT_IDX] != 4) or \
                        (output[POSZ_IDX] == 1100 and output[SLOT_IDX] != 5) or \
                        (output[POSZ_IDX] == 1300 and output[SLOT_IDX] != 6) or \
                        (output[POSZ_IDX] == 1500 and output[SLOT_IDX] != 7):
                    print("ERROR in line %d : %s" % (counter, line))
            else:
                if line[0] != '#':
                    print("Line not recognized : %s" % line)
